"""Functions that apply general operations on data."""

import matplotlib.pyplot as plt
import numpy as np
import pint
import pint_xarray
import xarray as xr
from IPython.display import clear_output
from numpy.fft import fft2, fftshift, ifft2
from scipy.ndimage import gaussian_filter
from skimage.registration import phase_cross_correlation

from peaks.core.fitting.models import _shirley_bg
from peaks.core.metadata.metadata_methods import compare_metadata
from peaks.core.process.fermi_level_correction import _flatten_EF
from peaks.core.utils.datatree_utils import get_list_of_DataArrays_from_DataTree
from peaks.core.utils.interpolation import _fast_bilinear_interpolate_rectilinear
from peaks.core.utils.misc import analysis_warning, dequantify_quantify_wrapper

ureg = pint_xarray.unit_registry


def norm(data, dim=None, **kwargs):
    """Function to apply a normalisation to data.

    Parameters
    ------------
    data : xarray.DataArray
        The data to be normalised.

    dim : str, optional
         Normalise data by an integrated DC along direction defined by dim, e.g. dim='eV' would normalise the data by an
         integrated MDC. Set dim='all' to normalise by the mean of the data. Defaults to None where the data is
         normalised to unity (i.e. normalised by the maximum value). Takes precedence over keyword arguments.

    **kwargs : slice, optional
        Slice to normalise by. E.g. eV=slice(105, 105.1) normalises by an integrated MDC defined by the eV slice given.
        Multiple slices can be defined to define a ROI to normalise by.

    Returns
    ------------
    norm_data : xarray.DataArray
        The normalised data.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # Load data
        disp = pks.load('disp.ibw')

        # Normalise the dispersion to unity
        disp_norm = disp.norm()

        # Normalise the dispersion by an integrated MDC
        disp_norm = disp.norm('eV')

        # Normalise the dispersion by an integrated EDC
        disp_norm = disp.norm('theta_par')

        # Normalise the dispersion by an EDC slice in the background
        disp_norm = disp.norm(theta_par=slice(15, 15.5))

        # Normalise by a region in the backgroun
        disp_norm = disp.norm(eV=slice(105, 105.1), theta_par=slice(-12, -8))

    """

    # Copy the input data to prevent overwriting issues
    norm_data = data.copy(deep=True)

    # If a dim has been provided, either normalise by the mean value of the data, or by an integrated DC
    if dim:
        # If dim is 'all', normalise by the mean value of the data
        if dim == "all":
            # Normalise by the mean value of the data
            norm_mean = norm_data.mean()
            norm_data /= norm_mean

            # Update analysis history
            norm_data.history.add(f"Data normalised by its mean value of {norm_mean}")

        # If not, normalise data by integrated DC along specified dim
        else:
            # Ensure dim is a valid dimension
            if dim not in norm_data.dims:
                raise Exception(
                    "{dim} is not a valid dimension of the inputted DataArray.".format(
                        dim=dim
                    )
                )

            # Normalise data by integrated DC along dim
            int_data = norm_data.mean(dim)
            norm_data /= int_data

            # Update analysis history
            norm_data.history.add(
                "Data normalised an integrated DC along {dim}".format(dim=dim)
            )

    # If no dim and no *kwargs have been provided, normalise to unity
    elif not dim and not kwargs:
        # Normalise data to unity
        norm_data /= norm_data.max()

        # Update analysis history
        norm_data.history.add("Data normalised to unity")

    # If kwargs have been provided, normalise by the slice defined in kwargs
    else:
        norm_slice = norm_data.sel(kwargs).mean(list(kwargs))
        norm_data /= norm_slice

        # Update analysis history
        norm_data.history.add(
            f"Data normalised by an integrated DC defined by the slice {slice}"
        )

    return norm_data


def bgs(
    data,
    subtraction=None,
    num_avg=1,
    offset_start=0,
    offset_end=0,
    max_iterations=10,
    **kwargs,
):
    """Function to subtract a background from data.

    Parameters
    ------------
    data : xarray.DataArray
        The data from which a background will be subtracted.

    subtraction : pint.Quantity, int, float, str
        The type of background to subtract:

            Set to a int/float to subtract that number, in the same units as the data e.g. subtraction=3.4. Pass a
            pint.Quantity to subtract a value with units, e.g. subtraction=3.4*pks.ureg('kcount/s').

            Set to 'all' to subtract the mean of the data.

            Set to a str of a valid dimension to subtract an integrated DC along the direction defined by the dimension,
            e.g. subtraction='eV' would subtract an integrated MDC.

            Set to 'Shirley' to subtract a Shirley background, e.g. subtraction='Shirley'. Additional arguments num_avg,
            offset_start, offset_end and max_iterations can be defined to optimise the Shirley background.

            Takes precedence over keyword arguments.

    num_avg : int, optional
        Shirley background optimisation parameter. The number of points to consider when calculating the average value
        of the data start and end points. Useful for noisy data. Defaults to 1.

    offset_start : float, optional
        Shirley background optimisation parameter. The offset to subtract from the data start value. Useful when data
        range does not completely cover the start (left) tail of the peak. Defaults to 0.

    offset_end : float, optional
        Shirley background optimisation parameter. The offset to subtract from the data end value. Useful when data
        range does not completely cover the end (right) tail of the peak. Defaults to 0.

    max_iterations : int, optional
        Shirley background optimisation parameter. The maximum number of iterations to allow for convergence of Shirley
        background. Defaults to 10.

    **kwargs : slice, optional
        Slice to define background for subtraction by. E.g. eV=slice(105, 105.1) subtracts an integrated MDC defined
        by the eV slice given. Multiple slices can be defined to define a ROI to subtract the mean of.

    Returns
    ------------
    bgs_data : xarray.DataArray
        The background subtracted data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp1.xy')

        S2p_XPS = load('XPS1.ibw').DOS()

        # Subtract a constant background of 3.5 from the data
        bgs_disp = disp.bkg(3.5)

        # Subtract a constant background equal to the mean value of the data from the data
        disp_norm = disp.bkg('all')

        # Subtract a background equal to an integrated MDC from the data
        disp_norm = disp.bkg('eV')

        # Subtract a background equal to an integrated EDC from the data
        disp_norm = disp.bkg('theta_par')

        # Subtract a Shirley background from the data
        bgs_S2p_XPS = S2p_XPS.bkg('Shirley')

        # Subtract a Shirley background from the data, using 3 points to calculate the average value of the data start
        # and end points
        bgs_S2p_XPS = S2p_XPS.bkg('Shirley', num_avg=3)

    """

    # Check a subtraction argument has been inputted
    if not subtraction and not kwargs:
        raise Exception(
            "No background subtraction argument has been inputted. Please specify either via the subtraction"
            " argument or by defining at least one slice in the kwargs with a dimension keyword."
        )
    # Copy the input data to prevent overwriting issues
    bgs_data = data.copy(deep=True)

    # If subtraction is an int or float, perform a constant background subtraction by a user-defined amount
    if isinstance(subtraction, (int, float, pint.Quantity)):
        if not isinstance(subtraction, pint.Quantity):
            # Parse the units from the data
            subtraction = subtraction * data.data.units

        # Subtract a constant background from the data
        bgs_data -= subtraction

        # Update analysis history
        bgs_data.history.add(
            "A constant background of {val} has been subtracted from the data".format(
                val=str(subtraction)
            )
        )

    # If subtraction is a str, there a few options of what the desired operation is.
    elif isinstance(subtraction, str):
        # If subtraction is 'Shirley'
        if subtraction == "Shirley":
            if "eV" not in bgs_data.dims:
                raise ValueError(
                    "Shirley background subtraction can only be performed on data with an 'eV' dimension."
                )

            # Calculate the Shirley background using the function _Shirley
            units = bgs_data.data.units
            Shirley_bkg = _shirley_bg(
                bgs_data.DOS().pint.dequantify(),  # Mean over all non-energy dimensions
                num_avg=num_avg,
                offset_start=offset_start,
                offset_end=offset_end,
                max_iterations=max_iterations,
            )
            Shirley_bkg = xr.DataArray(
                Shirley_bkg, dims="eV", coords={"eV": bgs_data.coords["eV"]}
            ).pint.quantify(units)
            # Subtract the Shirley background from the data
            bgs_data -= Shirley_bkg

            # Update analysis history
            bgs_data.history.add(
                "A Shirley background has been subtracted from the data"
            )

        # If subtraction is 'all'
        elif subtraction == "all":
            # Subtract a constant background equal to the mean value of the data from the data
            bgs_data -= bgs_data.data.mean()

            # Update analysis history
            bgs_data.history.add(
                "A constant background equal to the mean value of the data has been subtracted from the data"
            )

        # If subtraction is a valid dimension of the data
        elif subtraction in bgs_data.dims:
            # Integrate data along dimension defined by subtraction
            int_data = bgs_data.mean(subtraction)

            # Subtract data by integrated DC
            bgs_data.data = (bgs_data - int_data).data

            # Update analysis history
            bgs_data.history.add(
                "A background equal to an integrated DC along {subtraction} has been subtracted from the data".format(
                    subtraction=subtraction
                )
            )

        # Inputted argument for subtraction is not valid. Raise an error
        else:
            raise Exception(
                f"Subtraction type {subtraction} is not a valid argument. Expected int, float, 'all', 'Shirley', or a "
                f"valid dimension of the inputted DataArray, or for the subtraction to be defined by keyword arguemnts."
            )

    elif subtraction is None and kwargs:
        # If kwargs have been provided, subtract data dedined by the slice defined in kwargs
        bg_slice = bgs_data.sel(kwargs).mean(list(kwargs))
        bgs_data -= bg_slice

        # Update analysis history
        bgs_data.history.add(
            f"Data subtracted by an integrated DC defined by the slice {slice}"
        )
    # Invalid data type for subtraction. Raise an error
    else:
        raise Exception(
            "Invalid data type for subtraction. Please pass a pint.Quantity, int, float, 'all', 'Shirley', or a "
            "valid dimension of the inputted DataArray to the `subtraction` argument, or supply valid slices as "
            "keyword arguemnts."
        )

    return bgs_data


def bin_data(data, binning=None, boundary="trim", **binning_kwargs):
    """Shortcut function to bin data. Thin wrapper around :class:`xarray.DataArray.coarsen` but also with updating
    analysis history.

    Parameters
    ------------
    data : xarray.DataArray
        The data to be binned.

    binning : int, optional
        Size of bins to apply to all dimensions. Defaults to None (dimension-specific binning_kwargs must be defined).
        Takes priority over use of binning_kwargs.

    boundary : str, optional
        Determines how to handle boundaries. Defaults to 'trim' where bins are trimmed to fit the data.
        Other options are 'exact' and 'pad'; see :class:`xarray.DataArray.coarsen` for more information.

    **binning_kwargs : int, optional
        Used to define dimension-specific binning in the format dim = bin_size, e.g. theta_par = 2.

    Returns
    ------------
    binned_data : xarray.DataArray
        The binned data.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        disp = pks.load('disp.ibw')

        # Bin dispersion using bin sizes of 2 in both dimensions.
        disp_binned1 = disp.bin_data(2)

        # Bin dispersion using a bin size of 2 along theta_par, and a bin size of 3 along eV.
        disp_binned2 = disp.bin_data(theta_par=2, eV=3)

    """

    # If binning argument is defined, apply same size bins to all dimensions
    if binning:
        # If binning argument is not an integer, raise an error
        if not isinstance(binning, int):
            raise Exception("Binning value must be an integer.")

        # Overwrite any binning_kwargs since binning takes priority
        binning_kwargs = {}

        # Loop through all dimensions and define bin size in binning_kwargs
        for dim in data.dims:
            binning_kwargs[dim] = binning

    # If binning argument is not defined, get dimension-specific bins from binning_kwargs
    else:
        # If no binning_kwargs are supplied, raise an error
        if len(binning_kwargs) == 0:
            raise Exception(
                "No binning parameters set. Define either binning or binning_kwargs arguments."
            )

        # Loop through all items in binning_kwargs to ensure dictionary keys are valid dimensions and dictionary values
        # are integers, if not raise an error.
        for dim in binning_kwargs:
            if dim not in data.dims:
                raise Exception(
                    "{dim} is not a valid dimension of the inputted DataArray.".format(
                        dim=dim
                    )
                )
            if not isinstance(binning_kwargs[dim], int):
                raise Exception("Binning values must be integers.")

    # Apply binning to data
    binned_data = data.coarsen(binning_kwargs, boundary=boundary).mean()

    # Update analysis history
    binned_data = binned_data.history.assign(
        f"Binned data using the bins: {binning_kwargs}, boundary={boundary}"
    )

    return binned_data


def bin_spectra(data, binning=2, boundary="trim"):
    """Shortcut to .bin_data where spectral dimensions (energy, angle or k along slit) are attempted to be
    automatically determined and binned with the factor set in binning. All other dimensions are left as they are.

    Parameters
    ------------
    data : xarray.DataArray
        The data to be binned.

    binning : int, optional
        The factor by which to bin the spectral dimensions. Defaults to 2.

    boundary : str, optional
        Determines how to handle boundaries. Defaults to 'trim' where bins are trimmed to fit the data.
        Other options are 'exact' and 'pad'; see :class:`xarray.DataArray.coarsen` for more information.

    Returns
    ------------
    binned_data : xarray.DataArray
        The binned data.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        disp = pks.load('disp.ibw')

        # Bin dispersion by a factor of 2 along the spectral dimensions
        disp_binned = disp.bin_spectra()

    See Also
    ------------
    :meth:`xarray.DataArray.bin_data`

    """
    # Get the spectral dimensions
    spectral_dims = set(data.dims).intersection({"eV", "theta_par", "k_par"})

    # If no spectral dimensions are found, raise an error
    if len(spectral_dims) == 0:
        raise Exception("No spectral dimensions found in the inputted DataArray.")

    # Define the binning_kwargs
    binning_kwargs = {dim: binning for dim in spectral_dims}

    # Apply binning to data
    binned_data = data.bin_data(**binning_kwargs, boundary=boundary)

    return binned_data


def smooth(data, **smoothing_kwargs):
    """Function to smooth data by applying a Gaussian smoothing operator.

    Parameters
    ------------
    data : xarray.DataArray
        The data to smooth.

    **smoothing_kwargs : pint.Quantity, float
        Axes to smooth over in the format axis=FWHM, where FWHM is the relevant FWHM of the Gaussian for convolution
        in this direction, e.g. eV=0.1*pks.ureg('eV'). If a float with no unit is passed, the FWHM is assumed to be in
        the same units as the axis.

    Returns
    ------------
    smoothed_data : xarray.DataArray
        The smoothed data.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        disp = pks.load('disp.ibw')

        EDC1 = disp.EDC()

        # Smooth the dispersion by Gaussian filters with FWHMs for the theta_par and eV axes of 0.5 deg and 0.2 eV
        disp_smooth = disp.smooth(theta_par=0.5, eV=0.2)

        # Smooth the EDC by a Gaussian filter with FWHM for the eV axis of 0.2 eV
        EDC1_smooth = EDC1.smooth(eV=0.1)

    """

    # Check that some axes to smooth over were passed
    if len(smoothing_kwargs) == 0:
        raise Exception("Function requires axes to be smoothed over to be defined.")

    # Copy the input data to prevent overwriting issues
    smoothed_data = data.copy(deep=True)

    # Ensure units, getting from data axis if not passed and transforming to axis units if passed as a pint Quantity
    smoothing_kwargs_with_units = {}
    axis_units = {}
    for key, dim in smoothing_kwargs.items():
        axis_unit = data[key].units
        if isinstance(axis_unit, str):
            axis_units[key] = ureg(axis_unit)
        elif not isinstance(axis_unit, (pint.Unit, type(None))):
            raise TypeError(
                "Axis units must be a pint.Unit, str or None, not {type}".format(
                    type=type(axis_unit)
                )
            )
        axis_units[key] = axis_unit
        if isinstance(dim, pint.Quantity):
            smoothing_kwargs_with_units[key] = dim.to(axis_unit)
        else:
            smoothing_kwargs_with_units[key] = dim * axis_unit

    # Define the start of the analysis history update string
    hist = "Data smoothed with the following FWHMs along given axes: "

    # Make the sigma array (used to store standard deviations) as zeros, then update using supplied definitions
    sigma = np.zeros(len(data.dims))

    # Iterate through coordinates and determine the standard deviations in pixels from the DataArray axis scaling
    for count, dim in enumerate(data.dims):
        if dim not in smoothing_kwargs:  # No broadening for this dimension
            sigma[count] = 0
        else:  # Determine broadening in pixels from axis scaling
            delta = (
                abs(data[dim].data[1] - data[dim].data[0]) * axis_units[dim]
            )  # Pixel size in relevant units for axis
            # Must convert smoothing factor from FWHM to standard deviation (in pixels)
            sigma_px = (
                np.round(smoothing_kwargs_with_units[dim] / delta) / 2.35482005
            )  # Coordinate sigma in pixels
            if not sigma_px.dimensionless:
                raise Exception(
                    "Units conversion error. Check passed arguments are compatible with the axis units."
                )
            sigma[count] = sigma_px.magnitude  # Update standard deviations array
            hist += (
                str(dim) + ": " + str(smoothing_kwargs_with_units[dim]) + ", "
            )  # Update analysis history string
            smoothing_kwargs.pop(
                dim
            )  # Remove this axis from smoothing_kwargs for consistency check later

    # Extract the raw DataArray data
    array = smoothed_data.data.magnitude

    # Apply gaussian convolution to raw DataArray data
    array_sm = gaussian_filter(array, sigma)

    # Update DataArray with smoothed data
    smoothed_data.data = array_sm

    # Check that all supplied smoothing_kwargs are used, giving a warning if not (this occurs if an axis does not exist)
    if len(smoothing_kwargs) != 0:
        analysis_warning(
            "Not all supplied axes are coordinates of DataArray: {coords} have been ignored.".format(
                coords=str(smoothing_kwargs)
            ),
            title="Analysis info",
            warn_type="danger",
        )

    # Update the analysis history
    smoothed_data.history.add(hist[:-2])

    return smoothed_data


@dequantify_quantify_wrapper
def rotate(data, rotation, **centre_kwargs):
    """Function to rotate 2D or 3D data around a given centre of rotation.

    Parameters
    ------------
    data : xarray.DataArray
        The data to be rotated.

    rotation : int, float
        The rotation angle in degrees.

    **centre_kwargs : float, optional
        Used to define centre of rotation in the format dim=coord,
        e.g. theta_par=1.2 sets the theta_par centre as 1.2.
        Default centre of rotation is (0, 0).
        If data is 3D, assumes that the eV dimension is not to be rotated. To specify a
        different behaviour or if there is no eV dimenison, must pass the centre_kwargs.


    Returns
    ------------
    rotated_data : xarray.DataArray
        The rotated data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        FM = load('FM.zip')

        FS1 = FM.FS(E=78.45, dE=0.01)

        # Rotate Fermi surface around a (0,0) centre of rotation by 13 degrees
        FS1_rotated_1 = FS1.rotate(13)

        # Rotate Fermi surface around a (theta_par=5, ana_polar=5) centre of rotation by 50 degrees
        FS1_rotated_2 = FS1.rotate(50, theta_par=5, ana_polar=5)

    """

    # If the rotation is a multiple of 360, no need to do anything
    if rotation % 360 == 0:
        return data

    # Check data is 2D or 3D
    if len(data.dims) not in [2, 3]:
        raise Exception("Function only acts on 2D or 3D data.")

    # Check user-defined centre of rotations
    for dim in centre_kwargs.keys():
        if dim not in data.dims:
            raise Exception(f"{dim} is not a valid dimension of the inputted DataArray.")

    # Get the relevant dimensions and centre of rotation
    if len(data.dims) == 2:
        rot_dims = data.dims
    elif "eV" in data.dims and "eV" not in centre_kwargs:
        rot_dims = [dim for dim in data.dims if dim != "eV"]
    else:
        rot_dims = list(centre_kwargs.keys())
    centres = [centre_kwargs.get(dim, 0) for dim in rot_dims]

    if len(rot_dims) != 2:
        raise Exception(
            "Cannot parse rotation dimensions. Please specify the rotation centres as \
`dim0=value0, dim1=value1` where dim0 and dim1 are the names of the relevant dimension."
        )

    # Prepare to interpolate data onto expanded coordinate grid determined by rotation
    def _rot_point(dim0, dim1, cen, angle):
        """Rotate a point around a centre while preserving xarray's (dim0, dim1) notation

        Parameters
        ------------
        dim0 : float or np.ndarray
            The first coordinate (typically y-axis).

        dim1 : float or np.ndarray
            The second coordinate (typically x-axis).

        cen : tuple
            The centre of rotation as (dim0_centre, dim1_centre).

        angle : float
            The angle of rotation in degrees.

        Returns
        ------------
        new_dim0, new_dim1 : tuple
            The rotated coordinates (dim0, dim1).
        """
        angle_r = np.radians(angle)
        c = np.cos(angle_r)
        s = np.sin(angle_r)

        new_dim1 = (
            cen[1]  # Center along dim1 (x-axis)
            - (s * (dim0 - cen[0]))  # Y transformation
            + (c * (dim1 - cen[1]))  # X transformation
        )
        new_dim0 = (
            cen[0]  # Center along dim0 (y-axis)
            + (c * (dim0 - cen[0]))  # Y transformation
            + (s * (dim1 - cen[1]))  # X transformation
        )

        return new_dim0, new_dim1  # Keep (dim0, dim1) order

    # Calculate the new limits of the rotated data
    dim0_range = np.array([data[rot_dims[0]].min(), data[rot_dims[0]].max()])
    dim1_range = np.array([data[rot_dims[1]].min(), data[rot_dims[1]].max()])
    corners = _rot_point(dim0_range[None, :], dim1_range[:, None], centres, rotation)

    # Define new coordinates for rotated data
    new_coord0 = np.arange(
        np.min(corners[0]),
        np.max(corners[0]),
        data[rot_dims[0]].data[1] - data[rot_dims[0]].data[0],
    )
    new_coord1 = np.arange(
        np.min(corners[1]),
        np.max(corners[1]),
        data[rot_dims[1]].data[1] - data[rot_dims[1]].data[0],
    )

    # Inverse transform to get the old co-ordinate values for the rotated data
    new_dim0_vals, new_dim1_vals = _rot_point(
        new_coord0[:, None], new_coord1[None, :], centres, -rotation
    )

    # Interpolate inputted data onto the expanded coordinate grids
    interpolated_data = (
        xr.apply_ufunc(
            _fast_bilinear_interpolate_rectilinear,
            new_dim0_vals,
            new_dim1_vals,
            data.coords[rot_dims[0]],
            data.coords[rot_dims[1]],
            data,
            input_core_dims=[
                [rot_dims[0], rot_dims[1]],
                [rot_dims[0], rot_dims[1]],
                [rot_dims[0]],
                [rot_dims[1]],
                rot_dims,
            ],
            output_core_dims=[rot_dims],
            vectorize=True,
            exclude_dims=set(rot_dims),
            dask="parallelized",
            keep_attrs=True,
        )
        .assign_coords(
            {
                rot_dims[0]: new_coord0,
                rot_dims[1]: new_coord1,
            }
        )
        .transpose(*data.dims)
    )
    # Ensure the name, units and attributes are retained
    interpolated_data.attrs = data.attrs.copy()

    interpolated_data.name = data.name
    for dim in rot_dims:
        if "units" in data.coords[dim].attrs:
            interpolated_data.coords[dim].attrs["units"] = data.coords[dim].attrs[
                "units"
            ]

    # Update analysis history
    interpolated_data.history.add(f"Rotated data by {rotation} degrees")

    return interpolated_data


def sym(data, flipped=False, fillna=True, **sym_kwarg):
    """Function which primarily applies a symmetrisation of data around a given axis.
    It can alternatively be used to simply flip data around a given axis.

    Parameters
    ------------
    data : xarray.DataArray
        The data to be symmetrised.

    flipped : bool, optional
        Whether to return the flipped data rather than the sum of the original and
        flipped data. Defaults to False.

    fillna : bool, optional
        Whether to fill NaNs with 0s. NaNs occur for regions where the original and
        flipped data do not overlap. If fillna=True, regions without overlap will appear
        with half intensity (since only one of the original or flipped data contributes).
        Defaults to True.

    **sym_kwarg : float, optional
        Axis to symmetrise about in the format axis=value, where value is coordinate
        value around which the symmetrisation is performed, passed as a float of
        pint.Quantity e.g. theta_par=1.4, eV=16.8*pks.ureg('eV').
        Defaults to eV=0.

    Returns
    ------------
    sym_data : xarray.DataArray
        The symmetrised (or simply flipped) data.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        disp = pks.load('disp.ibw')

        # Symmetrise the dispersion about theta_par=3
        disp_sym = disp.sym(theta_par=3)

        # Flip the dispersion about theta_par=3
        disp_sym = disp.sym(theta_par=3, flipped=True)

    """

    # Copy the input data to prevent overwrite issues
    sym_data = data.copy(deep=True)

    # Check that more than one sym_kwarg has not been passed. If so, raise an error
    if len(sym_kwarg) > 1:  # Check only called with single axis kwarg
        raise Exception("Function can only be called with single axis.")

    # If no axis has been provided in sym_kwarg, set to the default symmetrisation axis
    # and coordinate of eV=0
    if len(sym_kwarg) == 0:
        sym_kwarg = {"eV": 0}

    # Get provided axis and coordinate to perform symmetrisation around
    sym_axis = next(iter(sym_kwarg))
    sym_coord = sym_kwarg[sym_axis]

    # Check that provided axis is a valid dimension of inputted DataArray,
    # if not raise an error
    if sym_axis not in sym_data.dims:
        raise Exception(
            "Provided symmetrisation axis is not a valid dimension of inputted DataArray."
        )
    if isinstance(sym_coord, pint.Quantity):
        # If a pint.Quantity is passed, convert to the units of the axis
        sym_coord = sym_coord.to(sym_data[sym_axis].units).magnitude

    # Check if the symmetrisation coordinate is within the range of the inputted
    # DataArray, if not raise an error
    if sym_coord < min(sym_data[sym_axis].data) or sym_coord > max(
        sym_data[sym_axis].data
    ):
        raise Exception(
            f"Provided symmetrisation coordinate ({sym_axis}={sym_coord}) is not within \
the coordinate range of the inputted DataArray"
        )

    # Generate flipped axis DataArray which maps the original axis to the flipped axis
    flipped_axis_values = (2 * sym_coord) - sym_data[sym_axis].data
    flipped_axis_xarray = xr.DataArray(
        flipped_axis_values, dims=[sym_axis], coords={sym_axis: sym_data[sym_axis].data}
    )

    # Flip the inputted DataArray by interpolating it onto the flipped axis
    flipped_data = sym_data.pint.dequantify().interp({sym_axis: flipped_axis_xarray})
    if hasattr(sym_data.data, "units"):
        flipped_data = flipped_data.pint.quantify(sym_data.data.units)

    # Fill NaNs with 0s if requested
    if fillna:
        flipped_data = flipped_data.fillna(0)

    # If only the flipped data is requested
    if flipped:
        # Assign sym_data to just the flipped data
        sym_data = flipped_data
        # Update the analysis history
        sym_data.history.add(f"Flipped data about {sym_kwarg}")

    # If the full symmetrisation is requested
    else:
        # Sum the original and flipped data
        sym_data += flipped_data
        # Update the analysis history
        sym_data.history.add(f"Symmetrised data about {sym_kwarg}")

    return sym_data


@dequantify_quantify_wrapper
def sym_nfold(data, nfold, expand=True, fillna=True, **centre_kwargs):
    """Function to perform an n-fold symmetrisation of data around a centre coordinate.

    Parameters
    ------------
    data : xarray.DataArray
        The data to be symmetrised.

    nfold : int
        The rotation order.

    expand : bool, optional
        Whether to expand the coordinate grid to view all symmetrised data.

    fillna : bool, optional
        Whether to fill NaNs with 0s. When the data is rotated, the coordinate grid is
        expanded. As such there will be regions with NaNs on the new coordinate grid.
        Setting `fillna=False` means such regions remain NaNs, leading to a loss of data
        when the different rotations are summed during the symmetrisation. Setting
        `fillna=True` sets the NaNs to 0, but when the different rotations are summed
        during the symmetrisation, such regions are scaled to allow for a consistent
        intensity of the symmetrised data. Some NaNs will remain for regions of the new
        coordinate grid where there is no data. Defaults to True

    **centre_kwargs : float, optional
        Used to define centre of rotation used for the symmetrisation in the format
        `dim=coord`, e.g. `theta_par=1.2` sets the theta_par centre as 1.2.
        Default centre of rotation is (0, 0).
        If data is 3D, assumes that the eV dimension is not to be rotated. To specify a
        different behaviour or if there is no eV dimenison, must pass the centre_kwargs.

    Returns
    ------------
    sym_data : xarray.DataArray
        The symmetrised data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        FM1 = load('FM1.zip')
        FS1 = FM1.MDC(E=75.62, dE=0.02)

        # Perform a 3-fold symmetrisation of the Fermi surface around a (0,0) centre of
        # rotation
        FS1_sym = FS1.sym_nfold(nfold=3)

        # Perform a 3-fold symmetrisation of the Fermi surface around a
        # (theta_par=3, ana_polar=5) centre of rotation, restricting the coordinate grid
        # of the output to that of the input, and not replacing NaNs.
        FS1_sym = FS1.sym(theta_par=3, ana_polar=5, expand=False, fillna=False)

    """

    # Check data is 2D or 3D
    if len(data.dims) not in [2, 3]:
        raise Exception("Function only acts on 2D or 3D data.")

    # Check user-defined centre of rotations
    for dim in centre_kwargs.keys():
        if dim not in data.dims:
            raise Exception(f"{dim} is not a valid dimension of the inputted DataArray.")

    # Get the relevant dimensions
    if len(data.dims) == 2:
        rot_dims = data.dims
    elif "eV" in data.dims and "eV" not in centre_kwargs:
        rot_dims = [dim for dim in data.dims if dim != "eV"]
    else:
        rot_dims = list(centre_kwargs.keys())

    if len(rot_dims) != 2:
        raise Exception(
            "Cannot parse rotation dimensions. Please specify the rotation centres as \
`dim0=value0, dim1=value1` where dim0 and dim1 are the names of the relevant dimension."
        )

    # Get dimensions of inputted data
    dim0 = rot_dims[0]
    dim1 = rot_dims[1]

    # Determine the rotations that will be applied and summed to produce symmetrised data
    rotation_values = np.linspace(0, 360, nfold + 1)[0:-1]

    # Define lists to store rotated data, and coordinate limits
    rotated_data = []

    # Copy the input data to prevent overwrite issues
    data_to_be_symmetrised = data.copy(deep=True)

    # Perform the required rotations, determine coordinate limits of each rotated data
    for rotation in rotation_values:
        rotated_data.append(data_to_be_symmetrised.rotate(rotation, **centre_kwargs))

    # Determine the coordinate limits desired for the data
    if expand:
        dim0_min = np.min([np.min(data[dim0].data) for data in rotated_data])
        dim0_max = np.max([np.max(data[dim0].data) for data in rotated_data])
        dim1_min = np.min([np.min(data[dim1].data) for data in rotated_data])
        dim1_max = np.max([np.max(data[dim1].data) for data in rotated_data])
        dim0_values = np.arange(
            dim0_min,
            dim0_max,
            data_to_be_symmetrised[dim0].data[1] - data_to_be_symmetrised[dim0].data[0],
        )
        dim1_values = np.arange(
            dim1_min,
            dim1_max,
            data_to_be_symmetrised[dim1].data[1] - data_to_be_symmetrised[dim1].data[0],
        )
    else:
        dim0_values = data_to_be_symmetrised[dim0].data
        dim1_values = data_to_be_symmetrised[dim1].data

    # Define list to store interpolate rotated data
    interp_rotated_data = []

    # Interpolate rotated data onto extremal coordinate grid
    for i, entry in enumerate(rotated_data):
        entry = entry.pint.dequantify()
        if i == 0 and not expand:  # No interp needed for the first array in this case
            current_rotated_data = entry.pint.dequantify()
        else:
            current_rotated_data = (
                xr.apply_ufunc(
                    _fast_bilinear_interpolate_rectilinear,
                    dim0_values[:, None] * np.ones_like(dim1_values),
                    dim1_values[None, :] * np.ones_like(dim0_values)[:, None],
                    entry[dim0].data,
                    entry[dim1].data,
                    entry,
                    input_core_dims=[
                        [dim0, dim1],
                        [dim0, dim1],
                        [dim0],
                        [dim1],
                        [dim0, dim1],
                    ],
                    output_core_dims=[[dim0, dim1]],
                    vectorize=True,
                    exclude_dims=set([dim0, dim1]),
                    dask="parallelized",
                    keep_attrs=True,
                )
                .assign_coords(
                    {
                        rot_dims[0]: dim0_values,
                        rot_dims[1]: dim1_values,
                    }
                )
                .transpose(*data.dims)
            )

        # If fillna=True, fill NaNs with 0 and make/update a  NaN_counter so that we can
        # rescale the data accordingly later
        if fillna:
            current_rotated_data = current_rotated_data.fillna(0)
            current_NaN_counter = (
                current_rotated_data.pint.dequantify().where(
                    current_rotated_data == 0, -1
                )
                + 1
            )  # 1 where a nan has been filled, as 0 elsewhere
            if i == 0:
                NaN_counter = current_NaN_counter
            else:
                NaN_counter += current_NaN_counter

        interp_rotated_data.append(current_rotated_data)

    # Sum the rotated data to get the symmetrised data
    sym_data = _sum_or_subtract_data(interp_rotated_data)

    # Remove sum_data analysis_history entry, and clear warning
    # (given during sum_data since analysis_history attrs will differ)
    del sym_data._analysis_history.records[-1]
    clear_output()

    # Sort out naming and units
    sym_data.name = f"{data.name}-sym-{nfold}-fold"
    for dim in rot_dims:
        if "units" in data.coords[dim].attrs:
            sym_data.coords[dim].attrs["units"] = data.coords[dim].attrs["units"]

    # Rescale data according to the NaN_counter so regions are of consistent intensity
    if fillna:
        sym_data.data = (sym_data / (NaN_counter.max() - NaN_counter).data).data

    # Update analysis history
    sym_data.history.add(
        f"Symmetrised data using a {nfold}-fold rotation about centre \
{dim0}={centre_kwargs.get(dim0, 0)}, {dim1}={centre_kwargs.get(dim1, 0)}"
    )

    return sym_data


@dequantify_quantify_wrapper
def degrid(data, width=0.1, height=0.1, cutoff=4):
    """Function which removes a mesh grid from 2D data by filtering its fast Fourier transform (FFT).

    Parameters
    ------------
    data : xarray.DataArray
        The 2D data to remove a mesh grid from.

    width : float, optional
        The width (as a fraction of the total width) of the central low-frequency region of the FFT (which defines most
        of the important data information) that is excluded from filtering analysis. Defaults to 0.1 (i.e 10 %).

    height : float, optional
        The height (as a fraction of the total height) of the central low-frequency region of the FFT (which defines
        most of the important data information) that is excluded from filtering analysis. Defaults to 0.1 (i.e. 10 %).

    cutoff : int, float, optional
        Data points (outside the central low-frequency region) of the FFT with an intensity larger than cutoff
        multiples of the mean of data (i.e. intensity > cutoff * (data).mean()) are removed. Defaults to 4.

    Returns
    ------------
    degrid_data : xarray.DataArray
        The data with the mesh grid removed.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM1.ibw')
        disp = SM.tot(spatial_int=True)

        # Remove the mesh grid from the dispersion using the default settings
        degrid_disp = disp.degrid()

        # Remove the mesh grid from the dispersion, excluding a larger central low-frequency region and considering
        # points above a lower cutoff intensity
        degrid_disp = disp.degrid(width=0.2, height=0.2, cutoff=3.2)

    """

    # Ensure data is a 2D scan, if not raise an error
    if len(data.dims) == 2:
        # Copy the input data to prevent overwrite issues
        data_to_degrid = data.copy(deep=True)
    else:
        raise Exception("Inputted data must be a 2D scan.")

    # Perform fast Fourier transform (FFT) of data_to_degrid
    FFT = fft2(data_to_degrid.data)

    # Rearrange the FFT by shifting the zero-frequency component to the centre of the np.array (equivalent to changing
    # range from (0, 2 Pi) to (-Pi, Pi) since FFT is periodic according to period defined by grid size). Useful for
    # visualisation and data manipulation
    FFT = fftshift(FFT)

    # Define n1 and n2 values, the discrete coordinates in FFT space
    n1_values = np.linspace(0, FFT.shape[0] - 1, FFT.shape[0])
    n2_values = np.linspace(0, FFT.shape[1] - 1, FFT.shape[1])

    # Shift n1 and n2 values in accordance with fftshift
    if len(n1_values) % 2 == 0:
        n1_values -= (FFT.shape[0]) / 2
    else:
        n1_values -= (FFT.shape[0] - 1) / 2

    if len(n2_values) % 2 == 0:
        n2_values -= (FFT.shape[1]) / 2
    else:
        n2_values -= (FFT.shape[1] - 1) / 2

    # Represent the magnitude of the FFT as an xarray.DataArray for plotting/manipulation purposes.
    FFT_DataArray = xr.DataArray(
        abs(FFT), dims=("n1", "n2"), coords={"n1": n1_values, "n2": n2_values}
    )

    # Define the centre region that we want to leave untouched
    n1_centre = [-1 * FFT.shape[0] * width / 2, FFT.shape[0] * width / 2]
    n2_centre = [-1 * FFT.shape[1] * height / 2, FFT.shape[1] * height / 2]

    # Get FFT_DataArray with just the central low-frequency region (which defines most of the important information)
    centre_FFT_DataArray = (
        FFT_DataArray.sel(
            n1=slice(n1_centre[0], n1_centre[1]), n2=slice(n2_centre[0], n2_centre[1])
        )
        .interp({"n1": n1_values, "n2": n2_values})
        .fillna(0)
    )

    # Subtract the central low-frequency region from FFT_DataArray, leaving the not so important information which we
    # can filter
    centre_subtracted_FFT_DataArray = FFT_DataArray - centre_FFT_DataArray

    # Redefine cutoff as cutoff multiples of the mean of the FFT
    cutoff = float(FFT_DataArray.mean()) * cutoff

    # Create a filter where any points of centre_subtracted_FFT_DataArray for which the intensity is above the cutoff is
    # assigned to 0. All other points are assigned to 1
    FFT_filter = xr.DataArray(
        np.where(centre_subtracted_FFT_DataArray.data > cutoff, 0, 1),
        dims=("n1", "n2"),
        coords={"n1": n1_values, "n2": n2_values},
    )

    # Ensure n=0 components are unaffected by changing their values in the filter to 1
    FFT_filter[int(np.where(n1_values == 0)[0].item())] = 1
    FFT_filter[:, int(np.where(n2_values == 0)[0].item())] = 1

    # Define degrid_data, and assign its contents to the inverse FFT of the original FFT of the data, multiplied by the
    # filter to remove intense high frequency Fourier components
    degrid_data = data_to_degrid.copy(deep=True)
    degrid_data.data = abs(ifft2(FFT * FFT_filter.data))

    # Update the analysis history
    degrid_data.history.add(
        "Data has been degridded by removing intense high frequency Fourier components"
    )

    # Determine the grid that was removed by calculating the difference between the inputted and degridded data
    grid = data_to_degrid - degrid_data

    # Set up a subplot to plot the Fourier analysis
    fig, axes = plt.subplots(figsize=(15, 4), ncols=4)

    # Plot the magnitude of the FFT, and the central low-frequency region that is excluded from filtering analysis
    FFT_DataArray.plot(
        ax=axes[0], y="n2", robust=True, add_colorbar=False, cmap="binary"
    )
    axes[0].plot(
        [n1_centre[0], n1_centre[1]],
        [n2_centre[0], n2_centre[0]],
        c="yellow",
        linewidth=2,
    )
    axes[0].plot(
        [n1_centre[0], n1_centre[1]],
        [n2_centre[1], n2_centre[1]],
        c="yellow",
        linewidth=2,
    )
    axes[0].plot(
        [n1_centre[0], n1_centre[0]],
        [n2_centre[0], n2_centre[1]],
        c="yellow",
        linewidth=2,
    )
    axes[0].plot(
        [n1_centre[1], n1_centre[1]],
        [n2_centre[0], n2_centre[1]],
        c="yellow",
        linewidth=2,
    )
    axes[0].set_title("Data transformed to \nFourier space")

    # Plot the FFT with the central central low-frequency region removed
    centre_subtracted_FFT_DataArray.plot(
        ax=axes[1], y="n2", robust=True, add_colorbar=False, cmap="binary"
    )
    axes[1].set_title("Important low frequency data \nnot considered for filter")

    # Plot the filter
    FFT_filter.plot(ax=axes[2], y="n2", add_colorbar=False, cmap="binary_r")
    axes[2].set_title("Filter determined by \npoints above cutoff")

    # Plot the FFT multiplied by the filter (thus with intense high frequency Fourier components removed)
    (FFT_DataArray * FFT_filter).plot(
        ax=axes[3], y="n2", robust=True, add_colorbar=False, cmap="binary"
    )
    axes[3].set_title("Apply filter to data to \nremove points above cutoff")

    plt.tight_layout()

    # Set up a subplot to plot the results of the degridding
    fig, axes = plt.subplots(figsize=(15, 7), ncols=3)

    # Plot the inputted data
    data_to_degrid.plot(ax=axes[0], add_colorbar=False, cmap="binary")
    axes[0].set_title("Input data")

    # Plot the degridded data
    degrid_data.plot(ax=axes[1], add_colorbar=False, cmap="binary")
    axes[1].set_title("Degridded data")

    # Plot the difference between the inputted and degridded data, representing the mesh grid removed
    grid.plot(ax=axes[2], add_colorbar=False, cmap="binary")
    axes[2].set_title("Difference")

    plt.tight_layout()

    return degrid_data


def _sum_or_subtract_data(data, _sum=True, quiet=False):
    """Function to sum or subtract two or more DataArrays together, maintaining the metadata.
    If the metadata of the DataArrays differ, that of the first inputted DataArray will be used.
    If the coordinate grids of the DataArrays differ, all DataArrays will be interpolated onto the
    coordinate grid of the first inputted DataArray.
    If in subtract mode, only two DataArrays can be inputted.

    Parameters
    ------------
    data : list or xarray.DataTree
        Any number of :class:`xarray.DataArray` items to sum together, either passed as a list or
        as a tree containing the relevant dataarrays.

    _sum : bool, optional
        Whether to sum or subtract the inputted DataArrays. Defaults to True (sum).

    quiet : bool, optional
        Whether to suppress warnings. Defaults to False.

    Returns
    ------------
    summed_data : xarray.DataArray
        The single summed :class:`xarray.DataArray`.

    """
    # If data is a DataTree, convert to a list
    if isinstance(data, xr.DataTree):
        data = get_list_of_DataArrays_from_DataTree(data)

    # Get the number of inputted DataArrays
    num_data = len(data)

    if not _sum and num_data != 2:
        raise Exception(
            "Data subtraction only accepts two DataArrays or a DataTree with two leaves."
        )

    # Copy the first DataArray
    data_0_data = data[0].copy(deep=True)

    # Copy the attributes of the first DataArray
    data_0_attrs = data_0_data.attrs.copy()

    # Remove scan_name from the attributes, and assign scan name to data_0_name
    data_0_name = data_0_data.metadata.scan.name

    # Remove history from the attributes
    data_history = []
    data_history.append(data_0_attrs.pop("_analysis_history", "NONE"))

    # Variable used to use to store the summed DataArray data (will be updated with other DataArrays)
    summed_data = data_0_data.copy(deep=True)

    # Variable used to use to store the summed DataArray name (will be updated with other DataArrays)
    summed_name = data_0_name

    # Flags used to determine whether the analysis history is updated with a caution informing the user of an
    # attributes/coordinates mismatch between the inputted DataArrays
    attrs_warn_flag = False
    coords_warn_flag = False

    # Iterate through the rest of the inputted DataArrays and sum together
    for i in range(1, num_data):
        # Get current DataArray information
        current_data = data[i].copy(deep=True)
        current_name = current_data.metadata.scan.name

        # Remove history from the current DataArray attributes
        data_history.append(current_data.attrs.pop("_analysis_history", "NONE"))

        # Ensure that the dimensions of the current DataArray match those of the first DataArray, raise an error if not
        if current_data.dims != data_0_data.dims:
            raise Exception("Inputted DataArrays must have the same dimensions.")

        # Ensure that the coordinates of the current DataArray match those of the first DataArray. If not, interpolate
        # the current DataArray onto the coordinate grid of the first DataArray
        for dim in current_data.dims:  # Loop through dimensions
            # Check if the coordinates of the current dimension do not match that of the first DataArray
            if (len(current_data[dim]) != len(data_0_data[dim])) or not (
                current_data[dim].data == data_0_data[dim].data
            ).all():
                # Interpolate the current DataArray onto the current dimension coordinate grid of the first DataArray
                original_units = (
                    current_data.data.units
                    if hasattr(current_data.data, "units")
                    else None
                )  # Store units before interpolation
                current_data = current_data.interp({dim: data_0_data[dim]})
                if original_units is not None:
                    current_data.data = (
                        current_data.data * original_units
                    )  # Restore units after interpolation
                coords_warn_flag = True  # Update warning flag
                # Display warning informing the user of the interpolation
                warning_str = (
                    "The {dim} coordinates of scan {current_name} do not match those of scan {data_0_name}. "
                    "Interpolated scan {current_name} onto the {dim} coordinate grid of scan {"
                    "data_0_name}."
                ).format(dim=dim, current_name=current_name, data_0_name=data_0_name)
                analysis_warning(
                    warning_str, title="Analysis info", warn_type="danger", quiet=quiet
                )

        # Determine any attributes (including nested attributes) of the current DataArray that do not match the first DataArray
        mismatched_attrs = compare_metadata(data_0_data, current_data)
        mismatched_attrs.pop("scan", None)  # Remove individual scan attributes

        def dict_to_html_table(d):
            html = """
                <style>
                    table { border-collapse: collapse; }
                    td, th { padding: 2px 5px; margin: 0; border: 1px solid black; }
                </style>
                <table>
                """
            for key, value in d.items():
                if isinstance(value, dict):
                    if (
                        len(value) == 2
                        and "value1" in value.keys()
                        and "value2" in value.keys()
                    ):
                        nested_items = list(value.items())
                        nested_str = f"{nested_items[0][1]}&nbsp;&nbsp; || &nbsp;&nbsp;{nested_items[1][1]}"
                        html += f"<tr><td><strong>{key}</strong></td><td>{nested_str}</td></tr>"
                    else:
                        value = dict_to_html_table(value)
                        html += (
                            f"<tr><td><strong>{key}</strong></td><td>{value}</td></tr>"
                        )
                elif isinstance(value, pint.Quantity):
                    value = str(value)
                    html += f"<tr><td><strong>{key}</strong></td><td>{value}</td></tr>"
                else:
                    html += f"<tr><td><strong>{key}</strong></td><td>{value}</td></tr>"
            html += "</table>"
            return html

        formated_mismatched_str = dict_to_html_table(mismatched_attrs)

        # If any attributes (except scan name) of the current DataArray do not match the first DataArray, display
        # a warning telling the user that the attributes of the first DataArray will be saved
        if len(mismatched_attrs) > 0:
            attrs_warn_flag = True  # Update warning flag
            warning_str = (
                f"The following attributes of scan {data_0_name} do not match those of scan {current_name}: "
                f"{formated_mismatched_str} Attributes of scan {data_0_name} kept."
            )
            analysis_warning(
                warning_str, title="Analysis info", warn_type="danger", quiet=quiet
            )

        # Add the current DataArray to the running summed total
        if _sum:
            summed_data += current_data.data
            summed_name += " + {current_name}".format(current_name=current_name)
        else:
            summed_data -= current_data.data
            summed_name += " - {current_name}".format(current_name=current_name)

    # Update summed data scan name
    summed_data.metadata.scan.set("name", summed_name, add_history=False)
    summed_data.name = summed_name

    # Update the analysis history
    if _sum:
        hist_str = "{num_data} scans summed together.".format(num_data=num_data)
    else:
        hist_str = "Scans subtracted."
    if (
        attrs_warn_flag
    ):  # If the there is an attributes mismatch, append information to analysis history
        hist_str += (
            f" CAUTION: mismatch of some attributes - those of scan {data_0_name} kept"
        )
    if (
        coords_warn_flag
    ):  # If the there is a coordinates mismatch, append information to analysis history
        hist_str += (
            f" CAUTION: mismatch of some coordinates - interpolated data onto "
            f"scan {data_0_name} coordinate grid"
        )
    total_hist = {"record": hist_str}
    for i in range(len(data_history)):
        total_hist[f"original scan {i} analysis history"] = data_history[i]
    summed_data = summed_data.history.assign(total_hist)

    return summed_data


def sum_data(data, quiet=False):
    """Function to sum two or more DataArrays together, maintaining the metadata.
    If the metadata of the DataArrays differ, that of the first inputted DataArray will be used.
    If the coordinate grids of the DataArrays differ, all DataArrays will be interpolated onto the
    coordinate grid of the first inputted DataArray.

    Parameters
    ------------
    data : list or xarray.DataTree
        Any number of :class:`xarray.DataArray` items to sum together, either passed as a list or
        as a tree containing the relevant dataarrays.

    quiet : bool, optional
        Whether to suppress warnings. Defaults to False.

    Returns
    ------------
    summed_data : xarray.DataArray
        The single summed :class:`xarray.DataArray`.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # From individual dataarrays
        disp1 = load('disp1.ibw')
        disp2 = load('disp2.ibw')
        disp_sum = pks.sum_data([disp1, disp2])  # Sum the dispersions

        # From a DataTree, suppressing warnings of mismatched metadata
        dt = pks.load(['disp1.ibw', 'disp2.ibw'])
        disp_sum = pks.sum_data(dt, quiet=True)  # Sum the dispersions
    """

    return _sum_or_subtract_data(data, _sum=True, quiet=quiet)


def subtract_data(data, quiet=False):
    """Function to subtract two DataArrays together, maintaining the metadata.
    If the metadata of the DataArrays differ, that of the first inputted DataArray will be used.
    If the coordinate grids of the DataArrays differ, all DataArrays will be interpolated onto the
    coordinate grid of the first inputted DataArray.

    Parameters
    ------------
    data : list or xarray.DataTree
        Any number of :class:`xarray.DataArray` items to sum together, either passed as a list or
        as a tree containing the relevant dataarrays.

    quiet : bool, optional
        Whether to suppress warnings. Defaults to False.

    Returns
    ------------
    summed_data : xarray.DataArray
        The single summed :class:`xarray.DataArray`.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # From individual dataarrays
        disp1 = load('disp1.ibw')
        disp2 = load('disp2.ibw')
        disp_diff = pks.subtract_data([disp1, disp2])  # Calculate the difference

        # From a DataTree, suppressing warnings of mismatched metadata
        dt = pks.load(['disp1.ibw', 'disp2.ibw'])
        disp_diff = pks.subtract_data(dt, quiet=True)  # Calculate the difference
    """

    return _sum_or_subtract_data(data, _sum=False, quiet=quiet)


def merge_data(data, dim="theta_par", sel=None, offsets=None, hv_match_rounding=0):
    """Function to merge two or more DataArrays together along a given dimension.

    Parameters
    ------------
    data : list or xarray.DataTree
        Any number of N-dimensional DataArrays to merge together, or a :class:`xarray.DataTree` contining the
        data to merge.

    dim : str, optional
        The dimension to merge along. Defaults to 'theta_par'.

    sel : slice, optional
        Selection of DataArrays in the list data along the dimension defined in dim, e.g. setting dim='theta_par' and
        sel=slice(-9,8.5) would perform the selection .sel(theta_par=slice(-9,8.5)) on each DataArray in the list data.
        Defaults to slice(None,None).

    offsets : list, numpy.ndarray, float, int, optional
        The offsets along the dimension defined in dim that must be applied to the DataArrays in the list data. If
        offsets is of type list (or numpy.ndarray), len(offsets) must equal len(data), and the offsets will be applied
        to their corresponding DataArray in data. If offsets is of type float (or int), a list of offsets will be
        generated with evenly spaced offsets starting at 0, e.g. if offsets=12, a list with length equal to len(data)
        will be created of the form [0, 12, 24, ...]. Defaults to None where no offsets are applied.

    hv_match_rounding : int, optional
        The number of decimal places to round the hv coordinate of the DataArrays to
        when checking for duplicated photon energies if merging hv scans along the hv
        axis. Defaults to 0.

    Returns
    ------------
    merged_data : xarray.DataArray
        The single merged :class:`xarray.DataArray`.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # Load some dispersions into a list
        disps = []
        for i in range(4):
            disps.append(pks.load(f'disp{i+1}.ibw'))

        # Load some XPS scans into a list
        XPS_data = [pks.load(f'XPS_{i}.ibw') for i in range(1,3)]

        # Load some partial hv scans into a list
        hv_data = [pks.load(f'hv_{i}.ibw') for i in range(1,3)]

        # Merge the dispersions (measured at subsequent polar values with 10 degree intervals) along 'theta_par',
        # applying the offsets [0, 10, 20, 30] as a list
        merged_disp = pks.merge_data([disp1, disp2, disp3, disp4], offsets=[0, 10, 20, 30])

        # As above, but defining offsets=10 to produce the same result
        merged_disp = pks.merge_data([disp1, disp2, disp3, disp4], offsets=10)

        # As above, but cutting the detector edges of the data by slicing theta_par between -9 and 9 to obtain a better
        # merge result
        merged_disp = pks.merge_data([disp1, disp2, disp3, disp4], sel=slice(-9,9), offsets=10)

        # Merge the XPS scans (measured over different energy ranges)
        merged_XPS = pks.merge_data([XPS_1, XPS_2], dim='eV')

        # Merge the partial hv scans (measured over different photon energy ranges)
        merged_hv = pks.merge_data([hv_1, hv_2], dim='hv')

    """
    if sel is None:
        sel = slice(None, None)

    # If data is a DataTree, convert to a list
    if isinstance(data, xr.DataTree):
        data = get_list_of_DataArrays_from_DataTree(data)

    # Ensure dim is a valid dimension
    if dim not in data[0].dims:
        raise Exception(
            "{dim} is not a valid dimension of the inputted data.".format(dim=dim)
        )

    # Copy the input data to prevent overwriting issues, and perform the selection along dim defined by sel
    data_to_merge = []
    data_history = []
    for item in data:
        if isinstance(item, xr.core.dataarray.DataArray):
            data_to_merge.append(item.copy(deep=True).sel({dim: sel}))
            if "_analysis_history" in item.attrs:
                data_history.append(item.attrs["_analysis_history"])
            else:
                data_history.append("None")
        else:
            raise Exception("Data must be a list of xarray.DataArrays.")

    if dim == "theta_par":
        # Remove any curvature of the Fermi level
        flattened_EF = False
        for i, current_data in enumerate(data_to_merge):
            if isinstance(current_data.metadata.get_EF_correction(), dict):
                data_to_merge[i] = _flatten_EF(current_data)
                flattened_EF = True
        if flattened_EF:
            analysis_warning(
                "The Fermi level curvature has been removed one or more of the supplied scans.",
                "info",
                title="$E_F$ curvature correction",
            )

    # Apply offsets to inputted data if required
    if offsets:
        # If offsets is a list (or numpy.ndarray), ensure len(offsets) == len(data)
        if isinstance(offsets, list) or isinstance(offsets, np.ndarray):
            if not len(offsets) == len(data):
                raise Exception(
                    "If offsets is provided as a list (or numpy.ndarray), the length of offsets and data must match."
                )

        # If offsets is a float or int, make offsets a list of the form [0, offsets*1, offsets*2, ...]
        elif isinstance(offsets, float) or isinstance(offsets, int):
            offsets = [i * offsets for i in range(len(data))]

        # If offset is not of type list, numpy.ndarray, float or int, raise an error
        else:
            raise Exception(
                "Invalid type for offsets. Argument must be a list, numpy.ndarray, float or int."
            )

        # Loop through data, apply offsets, and maintain coordinate units if any are present
        for i, current_data in enumerate(data_to_merge):
            data_to_merge[i].coords[dim] = current_data.coords[dim] - offsets[i]
            try:
                data_to_merge[i].coords[dim].attrs["units"] = (
                    data[i].coords[dim].attrs["units"]
                )
            except KeyError:
                pass

    # Ensure data_to_merge is arranged in order of increasing coordinates along dim
    min_values = []
    for item in data_to_merge:
        min_values.append([item.coords[dim].data.min(), item])
    min_values.sort()
    data_to_merge = [entry[1] for entry in min_values]

    # Initially define merged_data as the first entry of data_to_merge, and extract the scan name
    merged_data = data_to_merge[0]
    scan_name = merged_data.metadata.scan.name

    # Loop through the remaining entries of data_to_merge, merge the data with merged_data (using the function
    # _merge_two_DataArrays), and update the scan name
    for current_data in data_to_merge[1:]:
        if dim == "hv":
            merged_data = _join_two_hv_scans(
                merged_data.pint.dequantify(),
                current_data.pint.dequantify(),
                hv_match_rounding,
            ).pint.quantify()
        else:
            merged_data = _merge_two_DataArrays(
                merged_data.pint.dequantify(), current_data.pint.dequantify(), dim
            ).pint.quantify()
        scan_name += " & " + current_data.metadata.scan.name

    # Update analysis history
    history_str = f"Merged {scan_name} along {dim} "
    if offsets is not None:
        history_str += f"with offsets {offsets} "
    if sel.start is not None or sel.stop is not None:
        history_str += f"with data cropped to {dim}={sel} "
    history_str = history_str.rstrip() + "."
    history_record = {"record": history_str}
    for i in range(len(data_history)):
        history_record[f"original scan {i} analysis history"] = data_history[i]
    merged_data.attrs.pop("_analysis_history", None)
    merged_data = merged_data.history.assign(history_record)

    return merged_data


def _merge_two_DataArrays(DataArray1, DataArray2, dim):
    """Function to merge two N-dimensional DataArrays together along a given dimension.

    Parameters
    ------------
    DataArray1 : xarray.DataArray
        The first DataArray to be merged, with the lowest coordinates along dim.

    DataArray2 : xarray.DataArray
        The second DataArray to be merged, with the highest coordinates along dim

    dim : str
        The dimension to merge along.

    Returns
    ------------
    merged_data : xarray.DataArray
        The single merged :class:`xarray.DataArray`.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        from peaks.core.process.tools import _merge_two_DataArrays

        disp1 = load('disp1.nc')
        disp2 = load('disp2.nc')

        # Merge the dispersions along 'theta_par'
        merged_disp = merge_data(disp1, disp2, 'theta_par')

    """

    # Ensure dims of the inputted DataArrays are the same
    if not DataArray1.dims == DataArray2.dims:
        raise ValueError("The dimensions of the inputted DataArrays do not match.")

    # Ensure dim is a valid dimension
    if dim not in DataArray1.dims:
        raise ValueError(
            "{dim} is not a valid dimension of the inputted data.".format(dim=dim)
        )

    # Copy the input data to prevent overwriting issues
    DataArray1 = DataArray1.copy(deep=True)
    DataArray2 = DataArray2.copy(deep=True)

    # Determine overlap region of the two DataArrays
    overlap_limits = (
        DataArray2.coords[dim].data.min(),
        DataArray1.coords[dim].data.max(),
    )

    # Define coordinate axis along dim for merged_data
    coord_step = abs(DataArray1.coords[dim][1] - DataArray1.coords[dim][0])
    coord_limits = (
        DataArray1.coords[dim].data.min(),
        DataArray2.coords[dim].data.max(),
    )
    coord_num_points = int((coord_limits[1] - coord_limits[0]) / coord_step)
    coord_values = np.linspace(coord_limits[0], coord_limits[1], coord_num_points)

    # Interpolate DataArrays onto new coordinate grid
    DataArray1 = DataArray1.interp({dim: coord_values}).fillna(0)
    DataArray2 = DataArray2.interp({dim: coord_values}).fillna(0)

    # The overlap region will now be slightly different due to the new coordinate system. Find the indexes of the new
    # overlap region
    overlap_limits_indexes = (
        (np.abs(coord_values - overlap_limits[0])).argmin(),
        (np.abs(coord_values - overlap_limits[1])).argmin(),
    )

    # Determine the total counts within the overlap region of the two DataArrays
    DataArray1_overlap_intensity = float(
        DataArray1.isel(
            theta_par=slice(overlap_limits_indexes[0], overlap_limits_indexes[1])
        ).sum()
    )
    DataArray2_overlap_intensity = float(
        DataArray2.isel(
            theta_par=slice(overlap_limits_indexes[0], overlap_limits_indexes[1])
        ).sum()
    )

    # Scale DataArray2 such that the total counts within the overlap region of the two DataArrays is equal
    ratio = DataArray1_overlap_intensity / DataArray2_overlap_intensity
    DataArray2.data *= ratio

    # Before we sum the two DataArrays, we want to apply linear intensity reductions in the overlap region so that we
    # gradually go from the DataArray1 spectrum to the DataArray2 spectrum, obtaining equal contributions in the center
    # of the overlap region. Achieve this by defining a scaling for DataArray1 (called DataArray1_scaling) which is 1
    # over the theta_par region consisting of just DataArray1 (left_region), linearly decreases from 1 to 0 over the
    # overlap region, and is 0 over the theta_par region consisting of just DataArray2 (left_region). An equivalent
    # scaling for DataArray2 (called DataArray2_scaling) is obtained as DataArray2_scaling = 1 - DataArray1_scaling
    left_region = np.ones(overlap_limits_indexes[0])
    overlap_region = np.linspace(
        1, 0, (overlap_limits_indexes[1] - overlap_limits_indexes[0] + 1)
    )
    right_region = np.zeros(len(coord_values) - overlap_limits_indexes[1] - 1)
    DataArray1_scaling = np.concatenate((left_region, overlap_region, right_region))
    DataArray2_scaling = 1 - DataArray1_scaling

    # Represent the scaling as DataArrays so that they are associated with the correct dimension (makes function
    # applicable to any N-dimensional data)
    DataArray1_scaling = xr.DataArray(
        DataArray1_scaling, dims=[dim], coords={dim: coord_values}
    )
    DataArray2_scaling = xr.DataArray(
        DataArray2_scaling, dims=[dim], coords={dim: coord_values}
    )

    # Normalise DataArray1 and DataArray2 by their respective scaling along dim
    DataArray1 *= DataArray1_scaling
    DataArray2 *= DataArray2_scaling

    # Sum DataArray1 and DataArray2
    merged_data = _sum_or_subtract_data([DataArray1, DataArray2])

    # Remove sum_data analysis_history entry
    del merged_data._analysis_history.records[-1]

    return merged_data


def _join_two_hv_scans(scan1, scan2, hv_match_rounding=0):
    """Join two hv scans into a single hv scan

    Parameters
    ----------
    scan1 : xarray.DataArray
        The first scan to join

    scan2 : xarray.DataArray
        The second scan to join

    hv_match_rounding : int, optional
        The number of decimal places to round the hv values to before checking for
        duplicates, default is 0
    """

    # Check for duplicated hv values
    duplicate_hv = set(np.round(scan1.hv.data, hv_match_rounding)).intersection(
        set(np.round(scan2.hv.data, hv_match_rounding))
    )
    if len(duplicate_hv) > 0:
        error_message = f"""Scans {scan1.name} and {scan2.name} appear to contain one \
or more duplicate photon energies: {[float(hv) for hv in duplicate_hv]}
            Pre-select before passing to join_hv_scans function or pass parameter \
`hv_match_rounding` with precision set to the desired level.
            """
        raise ValueError(error_message)

    # Check kinetic energy range of the scan is the same
    if np.abs(np.ptp(scan1.eV.data) - np.ptp(scan2.eV.data)) > 0.01:
        raise ValueError("Scans appear to have different kinetic energy ranges")

    # Calculate the KE offsets of the first scans from each set
    offset = scan2.eV[0] - scan1.eV[0]

    # Shift the KE_delta values of the second scan and remap the KE axis to the first
    shifted_scan2 = scan2.assign_coords(
        {"eV": scan1.eV, "KE_delta": scan2.KE_delta + offset}
    )

    # Concatenate the two scans
    joined_scan = xr.concat([scan1, shifted_scan2], dim="hv")
    joined_scan = joined_scan.sortby("hv")

    return joined_scan


def estimate_sym_point(data, dims=None, upsample_factor=100):
    """Function to estimate the centrepoint of data that should be symmetric about an axis or axes.
    Used for e.g. estimating normal emission of an ARPES scan.

    Parameters
    ------------
    data : xarray.DataArray
        The data to estimate the centrepoint of.

    dims : string, tuple of string, list of string, optional
        The dimensions to estimate the centrepoint along.
        Defaults to None, which estimates the centrepoint along all dims of the array

    upsample_factor : int, optional
        The upsample factor used in the phase cross correlation algorithm.
        Defaults to 100.

    Returns
    ------------
    centre : dict
        The estimated centrepoint of the data along the specified dims.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # Find the centre of the dispersion data along the theta_par dimension
        disp = pks.load('disp.ibw')
        centre = estimate_sym_point(disp, dims='theta_par')

        # Find the centre of a Fermi surface map
        FS = pks.load('FS.ibw')
        centre = estimate_sym_point(FS, dims=['theta_par', 'polar'])

    """

    # Ensure dims is of list type
    if isinstance(dims, str):
        dims = [dims]
    elif isinstance(dims, tuple):
        dims = list(dims)
    elif dims is None:
        dims = data.dims

    # Flip the data along the specified dimensions
    dim_numbers = [data.dims.index(keyword) for keyword in dims if keyword in data.dims]
    data_array = data.data
    data_array_flipped = np.flip(data_array, axis=tuple(dim_numbers))

    # Perform phase cross correlation to estimate the offsets between the arrays
    shift, _, _ = phase_cross_correlation(
        data_array, data_array_flipped, upsample_factor=upsample_factor
    )

    # Work out the offsets in real co-ordinates
    centre = {}
    for i, dim in enumerate(dims):
        coord = data.coords[dim].values
        coord_delta = coord[1] - coord[0]
        coord_midpoint = (coord[0] + coord[-1]) / 2
        midpoint = coord_midpoint + ((shift[dim_numbers[i]] / 2) * coord_delta)
        centre[dim] = midpoint

    return centre


def drift_correction(reference_data, moving_data, orig_pos=None, **kwargs):
    """Estimate new position to correct for drift between two spatial maps.

    Parameters
    ------------
    reference_data : xarray.DataArray
        The original data, should be a 1D or 2D DataArray

    moving_data : xarray.DataArray
        The new data which should be over the same relative range and be of the same dimensions as orig_data,
        but which can be over a different absolute range

    orig_pos : dict, optional
         Dictionary specifying the original positions

    **kwargs : optional
        Additional keyword arguments to be passed to :class:`skimage.registration.phase_cross_correlation` for
        the subpixel image registration

    Returns
    ------------
    shift : dict
        Dictionary of shifts required to register ``moving_data`` with ``reference_data``.
    new_pos : dict, optional
        If orig_pos supplied, dictionary of position in new_map corresponding to orig_pos in orig_map
    """

    reference_data = reference_data.squeeze()
    moving_data = moving_data.squeeze()
    if len(reference_data.shape) >= 3 or len(moving_data.shape) >= 3:
        raise ValueError("Supplied data has more than 2 dimensions. Select a 2D slice.")

    # Set default upsample factor
    kwargs.setdefault("upsample_factor", 10)

    # Calculate shifts in pixel space
    shifts, _, _ = phase_cross_correlation(
        reference_data.data, moving_data.data, **kwargs
    )

    # Convert to real axis units
    dim_shift = {}
    for i, shift in enumerate(shifts):
        dim = reference_data.dims[i]
        orig_coord = reference_data[dim].data
        new_coord = moving_data[dim].data
        dim_delta = orig_coord[1] - orig_coord[0]
        dim_shift[dim] = shift * dim_delta.data - (new_coord[0] - orig_coord[0])

    if orig_pos:
        new_pos = {dim: orig_pos[dim] - dim_shift[dim] for dim in orig_pos}
        return dim_shift, new_pos
    return dim_shift



import xarray as xr

from peaks.core.accessors.accessor_methods import register_lazy_accessor

# List of function names and the module they belong to
functions_to_register = {
    "display.plotting": [
        "plot_DCs",
        "plot_3d_stack",
        "plot_fit_test",
        "plot_nanofocus",
    ],
    "fileIO.data_saving": [
        "save",
    ],
    "fitting.fit": ["fit", "fit_gold", "estimate_EF"],
    "GUI.disp_panels.disp": ["disp"],
    "process.data_select": [
        "drop_nan_borders",
        "drop_zero_borders",
        "DC",
        "MDC",
        "EDC",
        "DOS",
        "tot",
        "radial_cuts",
        "extract_cut",
        "mask_data",
        "disp_from_hv",
    ],
    "process.differentiate": [
        "deriv",
        "d2E",
        "d2k",
        "dEdk",
        "dkdE",
        "curvature",
        "min_gradient",
    ],
    "process.k_conversion": ["k_convert"],
    "process.tools": [
        "norm",
        "bgs",
        "bin_data",
        "bin_spectra",
        "smooth",
        "rotate",
        "sym",
        "sym_nfold",
        "degrid",
        "estimate_sym_point",
    ],
}


# Register each function as a lazy accessor on DataArray
for module_name, func_names in functions_to_register.items():
    for func_name in func_names:
        accessor_property = register_lazy_accessor(
            func_name, f"peaks.core.{module_name}", xr.DataArray
        )



try:
    import ase
except ImportError as e:
    raise ImportError(
        "The 'ase' module is required to use this module. Please install it using \
'pip install ase'."
    ) from e
import ase.dft.bz
import matplotlib.pyplot as plt
import numpy as np
import trimesh
from ase.dft.bz import bz_vertices
from scipy.spatial import ConvexHull
from scipy.spatial.transform import Rotation

from peaks.bz.utils import _get_atoms_cell_lattice


class LinearScalingTransform:
    """
    Linear scaling transformation for scaling coordinates along each axis."""

    def __init__(self, scaling_factors):
        """
        Initialize the linear scaling transform.

        Parameters
        ----------
        scaling_factors : tuple of float
            Scaling factors for each axis.
        """
        self.scaling_factors = scaling_factors

    def apply(self, coords):
        """
        Apply the linear scaling transformation.

        Parameters
        ----------
        coords : np.ndarray
            Input coordinates to be transformed.

        Returns
        -------
        transformed_coords : np.ndarray
            Transformed coordinates.
        """
        return coords * self.scaling_factors


def plot_bz(
    structure_or_lattice,
    surface=None,
    path="",
    special_points=None,
    vectors=False,
    azim=36,
    elev=30,
    scale=1,
    rotate=0,
    repeat=(1, 1, 1),
    ax=None,
    show=False,
    show_axes=None,
    **kwargs,
):
    """Plot the brillouin zone. Wrapper around ase.cell.BravaisLattice.plot_bz.

    Parameters
    ----------
    structure_or_lattice : ase.Atoms | ase.lattice.BravaisLattice
        The crystal structure or Bravais lattice representation of the material.
    surface : Tuple[int, int, int] | None, optional
        Surface orientation. Default is None, in which case the bulk BZ is plotted.
        To plot BZ corresponding to a specific surface, provide the miller indices as a
        tuple, e.g. `surface=(1, 1, 0)`.
    path : str | None, optional
        Path to show on the plot. Default is an empty string, in which case no path is
        shown. To show a default path for a DFT calculation, set `path=None`. To show a
        custom path, provide a string with the k-points, where different paths can be
        separated by a comma, e.g. `path=GZR,MX`. If no special_points are provided, the
        path must use only standard high-symmetry points in the BZ. If special_points
        are provided, the path can only use the special points specified.
    special_points : dict | None, optional
        Dictionary of special points to label on the BZ. Default is None, in which case
        standard special points are plotted. To plot specific points, pass a dictionary
        of the form `{"point_label": [kx, ky, kz]}`. To plot no special points, pass an
        empty dictionary `{}`.
    vectors : bool, optional
        if True, show the vector.
    azim : float | None, optional
        Azimuthal angle in degrees for viewing 3D BZ.
    elev : float | None, optional
        Elevation angle in degrees for viewing 3D BZ.
    scale : float | int, optional
        Scale the BZ. Defaults to 1, which gives the size in inv. Angstrom.
    rotate : float | int | list[float, float, float], optional
        Rotate the BZ by angle (in degrees). Defaults to 0. If a float or int is passed,
        the BZ is rotated around the z-axis by the specified angle. If a list of three
        floats is passed, the BZ is rotated by that rotation vector.
        For the vector rotation, see also :class:`scipy.spatial.transform.Rotation`.
    repeat: Tuple[int, int] | Tuple[int, int, int], optional
        Set the repeating draw of BZ. default is (1, 1, 1), no repeat.
    ax : Axes | Axes3D, optional
        matplolib Axes (Axes3D in 3D) object
    show : bool, optional
        If true, show the figure.
    show_axes : bool, optional
        If True, show the axes, if False, hide the axes, if None default to behaviour of
        existing plot, or hide the axes if no plot exists.
    **kwargs
        Additional keyword arguments to pass to ax.plot




    """

    atoms, cell, lattice = _get_atoms_cell_lattice(structure_or_lattice)

    if surface is not None:
        if not isinstance(surface, (tuple, list, np.ndarray)) and len(surface) != 3:
            raise ValueError(
                "Invalid input for surface orientation. Must be a tuple, \
list, or np.ndarray of length 3."
            )
        surface_atoms = ase.build.surface(atoms, surface, 1)
        lattice = surface_atoms.cell.get_bravais_lattice()

    # Determine the rotation and scaling transformations
    # Rotation
    if isinstance(rotate, (float, int)):
        rotate = [0, 0, rotate]
    elif not isinstance(rotate, (list, np.ndarray)):
        raise ValueError("Invalid input for rotate. Must be a float, int, or list.")
    r = Rotation.from_rotvec(rotate, degrees=True)

    # Scaling
    if not isinstance(scale, (float, int)):
        raise ValueError("Invalid input for scale. Must be a float or int.")
    # Default from ase is BZ in units of 2pi, so we need to scale it by 2pi
    scale *= 2 * np.pi
    scaling_transform = LinearScalingTransform([scale, scale, scale])

    transforms = [r, scaling_transform]
    lattice.plot_bz(
        path=path,
        special_points=special_points,
        vectors=vectors,
        azim=np.radians(azim),
        elev=np.radians(elev),
        transforms=transforms,
        repeat=repeat,
        ax=ax,
        show=show,
        **kwargs,
    )
    if ax:
        show_axes = show_axes if show_axes is not None else True
    ax = ax or plt.gca()
    if show_axes is not None:
        ax.axis("on" if show_axes else "off")


def plot_bz_section(
    structure_or_lattice,
    plane_origin=None,
    plane_normal=None,
    repeat=1,
    ax=None,
    show=False,
    show_axes=None,
    **kwargs,
):
    """Plot a cross section through the bulk Brillouin zone.

    Parameters
    ----------
    structure_or_lattice : ase.Atoms | ase.lattice.BravaisLattice
        The crystal structure or Bravais lattice representation of the material.
    plane_origin : list | np.ndarray, optional
        Origin of the plane. Default is [0, 0, 0].
    plane_normal : list | np.ndarray, optional
        Normal vector to the plane. Default is [0, 0, 1].
    repeat : int, optional
        Number of times to repeat the Brillouin zone in each direction.
    ax : Axes, optional
        Matplotlib Axes object.
    show : bool, optional
        If True, show the figure.
    show_axes : bool, optional
        If True, show the axes, if False, hide the axes, if None default to behaviour of
        existing plot, or hide the axes if no plot exists.
    **kwargs
        Additional keyword arguments to pass to ax.plot
    """

    if plane_origin is None:
        plane_origin = [0, 0, 0]
    if plane_normal is None:
        plane_normal = [0, 0, 1]

    atoms, cell, lattice = _get_atoms_cell_lattice(structure_or_lattice)
    reciprocal_cell = atoms.cell.reciprocal() * 2 * np.pi  # in inv. Angstrom

    # Define the offsets
    offsets = np.array(
        [
            [i, j, k]
            for i in range(-repeat, repeat + 1)
            for j in range(-repeat, repeat + 1)
            for k in range(-repeat, repeat + 1)
        ]
    )
    offset_vectors = np.dot(offsets, reciprocal_cell)

    # Get the original BZ vertices
    bz_faces = bz_vertices(reciprocal_cell)
    vertices_original = np.vstack([face[0] for face in bz_faces])

    # Set up plot
    if not ax:
        fig, ax = plt.subplots()

    for offset in offset_vectors:
        vertices = vertices_original + offset

        # Compute convex hull for the 3D Brillouin zone
        hull = ConvexHull(vertices)
        mesh = trimesh.Trimesh(vertices=vertices, faces=hull.simplices)

        # Compute intersection
        slice_ = mesh.section(plane_origin=plane_origin, plane_normal=plane_normal)

        if slice_ is not None:
            # Extract absolute intersection vertices
            slice_vertices = slice_.vertices

            # Find two orthonormal basis vectors orthogonal to the plane normal
            plane_normal = np.array(plane_normal)
            u, _, vh = np.linalg.svd(plane_normal.reshape(1, 3))
            basis = vh[1:3]  # Two orthonormal vectors perpendicular to plane_normal

            # Project slice vertices onto the 2D basis
            xy = slice_vertices @ basis.T

            # Robust 2D Convex Hull calculation
            hull_2D = ConvexHull(xy)
            hull_xy = xy[np.append(hull_2D.vertices, hull_2D.vertices[0])]

            # Plot the hull
            ax.plot(hull_xy[:, 0], hull_xy[:, 1], color="k", **kwargs)

    if show_axes is not None:
        ax.axis("on" if show_axes else "off")

    if show:
        plt.show()



try:
    import ase
except ImportError as e:
    raise ImportError(
        "The 'ase' module is required to use this module. Please install it using \
'pip install ase'."
    ) from e

import numpy as np
import pandas as pd


def _get_atoms_cell_lattice(structure_or_lattice):
    """Get atoms, cell, and lattice from the given structure or lattice.

    Parameters
    ----------
    structure_or_lattice : ase.Atoms or ase.lattice.BravaisLattice
        Input structure or lattice object.

    Returns
    -------
    tuple(ase.Atoms, ase.cell.Cell, ase.lattice.BravaisLattice)
        Tuple containing the atoms, cell, and lattice.
    """
    if isinstance(structure_or_lattice, ase.Atoms):
        atoms = structure_or_lattice
        lattice = structure_or_lattice.cell.get_bravais_lattice()
        cell = structure_or_lattice.cell
    elif isinstance(structure_or_lattice, ase.lattice.BravaisLattice):
        lattice = structure_or_lattice
        cell = lattice.tocell()
        atoms = ase.Atoms(cell=cell, pbc=True)
    else:
        raise ValueError(
            "Invalid input. Please provide an ase.Atoms or an ase.lattice.BravaisLattice\
 object."
        )

    return atoms, cell, lattice


def _get_special_points_inv_Ang(lattice):
    """
    Get the symmetry points for the Brillouin zone of the given lattice.

    Parameters
    ----------
    lattice : ase.lattice.BravaisLattice
        Input lattice object.

    Returns
    -------
    sym_points : dict
        Dictionary containing the symmetry points, with values in inverse Angstrom.
    """
    sym_points = lattice.get_special_points()
    sym_points_inv_Ang = {
        k: ase.dft.kpoints.kpoint_convert(lattice.tocell(), skpts_kc=v)
        for k, v in sym_points.items()
    }  # Convert to k in inv Angstrom

    return sym_points_inv_Ang


def sym_points(structure_or_lattice, surface=None, hv=None):
    """
    Get the symmetry points for the Brillouin zone of the given structure or lattice.

    Parameters
    ----------
    structure_or_lattice : ase.Atoms or ase.lattice.BravaisLattice
        Input structure or lattice object.
    surface : tuple | list | np.ndarray | None
        Surface orientation of the structure, optional (default=None --> bulk)
    hv : float | None
        Photon energy in eV, to be used to generate angle along slit for k-points
        Optional (default=None, no angle calculation)

    Returns
    -------
    sym_points : dict
        Dictionary containing the symmetry points
    """
    atoms, cell, lattice = _get_atoms_cell_lattice(structure_or_lattice)

    if surface is not None:
        if not isinstance(surface, (tuple, list, np.ndarray)) and len(surface) != 3:
            raise ValueError(
                "Invalid input for surface orientation. Must be a tuple, \
list, or np.ndarray of length 3."
            )
        surface_atoms = ase.build.surface(atoms, surface, 1)
        lattice = surface_atoms.cell.get_bravais_lattice()

    df = pd.DataFrame.from_dict(
        _get_special_points_inv_Ang(lattice),
        orient="index",
        columns=["k_x", "k_y", "k_z"],
    )
    df["|k|"] = df.apply(np.linalg.norm, axis=1)
    if hv is not None:
        df[f"angle_along_slit @ {hv} eV"] = np.round(
            np.degrees(np.arcsin(df["|k|"] / (0.5123 * np.sqrt(hv - 4.5)))), 1
        )

    return df



from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.fileIO.loc_registry import register_loader


@register_loader
class BaseStructureLoader(BaseDataLoader):
    """Base class for data loaders for .cif structure files."""

    _loc_name = "cif"

    @classmethod
    def _load(cls, fpath, lazy, metadata, quiet, **kwargs):
        """Load the structure data."""

        try:
            import ase.io
        except ImportError as e:
            raise ImportError(
                "The 'ase' module is required to use this module. Please install it using \
        'pip install ase'."
            ) from e

        return ase.io.read(fpath)



from datetime import datetime, timezone
from typing import Optional, Union

import h5py
import numpy as np
import pint
import pint_xarray
import xarray as xr
from dateutil.parser import parse

from peaks.core.accessors.accessor_methods import register_accessor
from peaks.core.fileIO.base_arpes_data_classes.base_arpes_data_class import (
    BaseARPESDataLoader,
)
from peaks.core.fileIO.base_data_classes.base_hdf5_class import BaseHDF5DataLoader
from peaks.core.fileIO.loc_registry import register_loader
from peaks.core.metadata.base_metadata_models import BaseMetadataModel, Quantity
from peaks.core.options import opts
from peaks.core.utils.misc import analysis_warning

ureg = pint_xarray.unit_registry


class DiamondNXSLoader(BaseHDF5DataLoader):
    """Helper class for parsing the NeXus file format used at Diamond Light Source."""

    _data_group_key_resolution_order = []
    _core_data_key_resolution_order = []

    @property
    def data_key_resolution_order(self):
        """The priority order of keys to specify the main data group in the file"""
        return self._data_group_key_resolution_order

    @staticmethod
    def get_root_key(f):
        """Get the root key of the file, which is assumed to be the only key in the file at root level

        Parameters
        ----------
        f : h5py.File
            The h5py file object (should be open).

        Returns
        -------
        root_key : str
            The root key of the file.
        """

        if len(f.keys()) != 1:
            raise ValueError(
                "The file has multiple root keys, which is not expected for the standard Diamond NeXus file."
            )
        return next(iter(f.keys()))

    @classmethod
    def get_data_group_addr(cls, f, root_key=None):
        """Get the main data group from the file, with priority as per _data_key_resolution_order

        Parameters
        ----------
        f : h5py.File
            The h5py file object (should be open).
        root_key : str, optional
            The root key of the file. If not provided, it is automatically determined as the only key at root level.

        Returns
        -------
        data_group_addr : str
            The full path of the main data group in the file.
        """

        if root_key is None:
            root_key = cls.get_root_key(f)

        # Get all keys within this root level
        scan_keys = f[root_key].keys()

        # Check for the main data group, with priority as per data_key_resolution_order
        data_group_key = None
        for key in cls._data_group_key_resolution_order:
            if key in scan_keys:
                data_group_key = key
                break
        else:
            # If no key from data_key_resolution_order is found, fall back to the first key in the group
            data_group_key = next(iter(scan_keys))

        data_group_addr = f"{root_key}/{data_group_key}"
        return data_group_addr

    @staticmethod
    def _parse_nxs_primary_attr(attr):
        """Parse the attribute value for a signal or primary key to a standard format"""
        if isinstance(attr, bytes):
            attr = attr.decode()

        if isinstance(attr, np.ndarray):
            if attr.size == 1:
                attr = attr.item()

        if isinstance(attr, str):
            if "," in attr:
                attr = [int(i) for i in attr.split(",")]
            else:
                attr = int(attr)
        elif isinstance(attr, list):
            attr = [int(i.decode()) if isinstance(i, bytes) else int(i) for i in attr]
            if len(attr) == 1:
                attr = attr[0]

        return attr

    @staticmethod
    def _parse_nxs_axis_attr(attr):
        """Parse the attribute value for an axis key to a standard format"""
        if isinstance(attr, (list, np.ndarray)):
            attr = attr[0]
        if isinstance(attr, bytes):
            attr = attr.decode()
        return attr.decode() if isinstance(attr, bytes) else attr

    @classmethod
    def get_core_data_key(cls, f, data_address):
        """Find the core data in an HDF5/nxs group. First attempts to resolve data by key located in
        `_core_data_key_resolution_order` class variable. If that fails, fall back to trying to automatically parse
        by looking for a signal attribute flag

        Parameters
        ----------
        f : h5py.File
            The h5py file object (should be open).

        data_address : str
            The key of the core data in the `data_address` group of the HDF5 file
        """

        # Parse all data entries with a signal key
        data_group = f[data_address]

        # Check for key in the resolution order
        for key in cls._core_data_key_resolution_order:
            if key in data_group.keys():
                return key

        # Otherwise fall back to looking for a signal attribute
        signal_data_mappings = {
            key: data_group[key].attrs.get("signal")
            for key in data_group.keys()
            if "signal" in data_group[key].attrs
        }

        core_data_key = None
        if len(signal_data_mappings) == 1:
            core_data_key = next(iter(signal_data_mappings.keys()))
        elif len(signal_data_mappings) > 1:
            # Take the primary data based on the resolution order of the @signal attr
            core_data_key = min(
                signal_data_mappings,
                key=lambda k: cls._parse_nxs_primary_attr(signal_data_mappings[k]),
            )

        if core_data_key:
            return core_data_key

        # If no signal attribute found, raise an error
        raise ValueError(
            f"None of the expected data keys in {cls._core_data_key_resolution_order} found in the data, "
            f"and no signal attribute found to mark primary data in the group."
        )


@register_loader
class I05ARPESLoader(DiamondNXSLoader, BaseARPESDataLoader):
    _loc_name = "Diamond_I05_ARPES"
    _loc_description = "HR-ARPES branch line of I05 beamline at Diamond Light Source"
    _loc_url = "https://www.diamond.ac.uk/Instruments/Structures-and-Surfaces/I05.html"
    _analyser_slit_angle = 0 * ureg.deg

    _manipulator_name_conventions = {
        # HR-branch
        "polar": "sapolar",
        "tilt": "satilt",
        "azi": "saazimuth",
        "x1": "sax",
        "x2": "saz",
        "x3": "say",
    }

    _manipulator_sign_conventions = {
        # HR-branch
        "tilt": -1,
        "azi": -1,
        "x1": -1,
    }

    _analyser_name_conventions = {
        "deflector_perp": "deflector_x",
        "deflector_parallel": "deflector_y",
        "eV": ["energies", "kinetic_energy_center"],
        "theta_par": "angles",
        "hv": ["energy", "value"],
    }

    _hdf5_metadata_key_mappings = {
        "scan_command": "entry1/scan_command",
        "manipulator_polar": "entry1/instrument/manipulator/sapolar",
        "manipulator_tilt": "entry1/instrument/manipulator/satilt",
        "manipulator_azi": "entry1/instrument/manipulator/saazimuth",
        "manipulator_x1": "entry1/instrument/manipulator/sax",
        "manipulator_x2": "entry1/instrument/manipulator/say",
        "manipulator_x3": "entry1/instrument/manipulator/saz",
        "manipulator_salong": "entry1/instrument/manipulator/salong",
        "analyser_model": lambda f: (
            (
                "FIXED_VALUE:MBS A1"
                if parse(
                    (
                        (f["entry1/start_time"][()]).decode()
                        if isinstance(f["entry1/start_time"][()], bytes)
                        else f["entry1/start_time"][()]
                    )
                )
                > datetime(2021, 7, 1, tzinfo=timezone.utc)
                else "FIXED_VALUE:Scienta R4000"
            )
            if "entry1/start_time" in f
            else None
        ),
        "analyser_slit_width": "entry1/instrument/analyser/entrance_slit_size",
        "analyser_slit_width_identifier": "entry1/instrument/analyser_total/entrance_slit_setting",
        "analyser_eV": "entry1/instrument/analyser/energies",
        "analyser_step_size": lambda f: (
            np.ptp(f["entry1/instrument/analyser/energies"])
            / (max(f["entry1/instrument/analyser/energies"].shape) - 1)
            if "entry1/instrument/analyser/energies" in f
            else None
        ),
        "analyser_PE": "entry1/instrument/analyser/pass_energy",
        "analyser_sweeps": [
            "entry1/instrument/analyser/number_of_iterations",
            "entry1/instrument/analyser/number_of_cycles",
            "entry1/instrument/analyser/number_of_frames",
        ],
        "analyser_dwell": "entry1/instrument/analyser/time_for_frames",
        "analyser_lens_mode": "entry1/instrument/analyser/lens_mode",
        "analyser_acquisition_mode": "entry1/instrument/analyser/acquisition_mode",
        "analyser_eV_type": "FIXED_VALUE:kinetic",
        "analyser_deflector_parallel": "entry1/instrument/deflector_y/deflector_y",
        "analyser_deflector_perp": "entry1/instrument/deflector_x/deflector_x",
        "temperature_sample": "entry1/sample/temperature",
        "temperature_cryostat": "entry1/sample/cryostat_temperature",
        "temperature_shield": "entry1/sample/shield_temperature",
        "temperature_setpoint": "entry1/sample/temperature_demand",
        "photon_hv": "entry1/instrument/monochromator/energy",
        "photon_polarisation": [
            "entry1/instrument/insertion_device/beam/final_polarisation_label",
            "entry1/instrument/insertion_device/beam/final_polarisation_label\n\t\t\t\t\t\t\t\t",
        ],
        "photon_exit_slit": "entry1/instrument/monochromator/exit_slit_size",
        "timestamp": "entry1/start_time",
    }

    _data_group_key_resolution_order = ["analyser"]
    _core_data_key_resolution_order = ["analyser"]

    @classmethod
    def _load_data(cls, fpath, lazy, **kwargs):
        """Load the data from a Diamond I05 ARPES NeXus file. Try to automatically determine the core scan structure."""

        with h5py.File(fpath, "r") as f:
            # Parse the data group structure
            data_group_addr = cls.get_data_group_addr(f)
            # Find the signal (i.e. core) data from the signal key
            core_data_key = cls.get_core_data_key(f, data_group_addr)

            # Sort remaining data entries based on these axis and primary keys
            dim_names = set(f[data_group_addr].keys()) - {core_data_key}
            dim_name_to_axis_mapping = {
                dim: cls._parse_nxs_axis_attr(
                    f[f"{data_group_addr}/{dim}"].attrs.get("axis")
                )
                for dim in dim_names
            }

            # Parse the possible dimensions for each axis, excluding current which is always a `metadata` entry
            axis_to_dim_name_mapping = {
                axis: [
                    dim
                    for dim, axis2 in dim_name_to_axis_mapping.items()
                    if (axis == axis2 and dim.lower() != "current")
                ]
                for axis in set(dim_name_to_axis_mapping.values())
            }
            # Sort these based on their primary attribute
            # Get rid of attributes that don't have an axis link
            axis_to_dim_name_mapping.pop(None, None)
            axis_to_dim_name_mapping_sorted = {}
            for axis, dims in axis_to_dim_name_mapping.items():
                axis_to_dim_name_mapping_sorted[axis] = sorted(
                    dims,
                    key=lambda k: int(
                        cls._parse_nxs_primary_attr(
                            f[f"{data_group_addr}/{k}"].attrs.get("primary", 99)
                        )
                    ),
                )

                # Handle edge cases ############################################
                # ana_polar scan via dummy motor (I05 Nano)
                if axis_to_dim_name_mapping_sorted[axis] == [
                    "dummy_motor",
                    "analyser_polar_angle",
                ]:
                    axis_to_dim_name_mapping_sorted[axis] = [
                        "analyser_polar_angle",
                        "dummy_motor",
                    ]
                # Secondary spatial axis in a polar map
                if "sapolar" in axis_to_dim_name_mapping_sorted[axis]:
                    other_dim = list(
                        set(axis_to_dim_name_mapping_sorted[axis]) - set(["sapolar"])
                    )
                    if len(other_dim) == 1 and other_dim[0] in ["sax", "say", "salong"]:
                        axis_to_dim_name_mapping_sorted[axis] = [
                            "sapolar",
                            other_dim[0],
                        ]
                # hv scan
                if "energy" in axis_to_dim_name_mapping_sorted[axis]:
                    other_dim = list(
                        set(axis_to_dim_name_mapping_sorted[axis]) - set(["energy"])
                    )
                    axis_to_dim_name_mapping_sorted[axis] = ["energy"]
                    axis_to_dim_name_mapping_sorted[axis].extend(other_dim)

            # Define the primary dims and coords
            data_ndim = f[f"{data_group_addr}/{core_data_key}"].ndim
            dim_mapping = {}
            coords = {}
            for axis, dims in axis_to_dim_name_mapping_sorted.items():
                if "," in axis:
                    axes = axis.split(",")
                    n_axes = len(axes)
                    for i in range(n_axes):
                        dim_mapping[axes[i]] = dims[i]
                        indexer = [0] * n_axes
                        indexer[i] = slice(None)
                        coords[dims[i]] = f[f"{data_group_addr}/{dims[i]}"][
                            tuple(indexer)
                        ]
                elif len(dims) > 0:
                    dim_mapping[axis] = dims[0]
                    coords[dims[0]] = f[f"{data_group_addr}/{dims[0]}"][()].squeeze()
            dims = [dim_mapping.get(str(i + 1), "dummy") for i in range(data_ndim)]
            units = {}
            for dim in dims:
                try:
                    units[dim] = f[f"{data_group_addr}/{dim}"].attrs.get("units")
                except KeyError:
                    pass
            units["spectrum"] = f[f"{data_group_addr}/{core_data_key}"].attrs.get(
                "units", "counts"
            )
            units = {
                key: (cls._parse_nxs_axis_attr(value)) for key, value in units.items()
            }

        # Load the core data in a way that supports lazy loading
        ds = xr.open_dataset(
            fpath,
            group=data_group_addr,
            engine="h5netcdf",
            phony_dims="sort",
            chunks="auto",
        )
        # Map local dimension names keeping I05 convention for now
        ds = ds.rename(
            {orig: new for orig, new in zip(ds[core_data_key].dims, dims, strict=True)}
        )

        # Get the core data
        da = ds[core_data_key]
        da.attrs = {}

        # Apply the coordinates
        coords_to_apply = {dim: coords.get(dim) for dim in dims if dim != "dummy"}
        # Handle the special case of an hv scan, where the kinetic energy is 2D
        for dim, coord in coords_to_apply.copy().items():
            if coord.ndim == 2 and not np.array_equal(
                coords_to_apply["energies"][0], coords_to_apply["energies"][1]
            ):
                # Should be a 2D array with shape (value, KE) where value is the changing hv dim
                hv_coord_label = {"value", "energy"}.intersection(
                    set(coords_to_apply.keys())
                ).pop()

                if coords_to_apply.get(hv_coord_label).shape[0] == coord.shape[0]:
                    coords_to_apply["KE_delta"] = (
                        hv_coord_label,
                        coord[:, 0] - coord[0, 0],
                    )
                    coords_to_apply[dim] = coord[0]
                else:
                    analysis_warning(
                        f"Struggling to parse dimensions of {dim}. If this is a hv scan, may need to "
                        f"manually add a 'KE_delta' dimension.",
                        "warning",
                        "Unexpected data shape",
                    )
            elif coord.ndim == 2 and np.array_equal(
                coords_to_apply["energies"][0], coords_to_apply["energies"][1]
            ):
                # Deal with deflector maps with the energy axis somehow to be 2D
                coords_to_apply[dim] = coord[0]
        da = da.assign_coords(coords_to_apply)

        # Normalise data by I0 if required
        if kwargs.get("norm_by_I0", False):
            if "current" in dim_name_to_axis_mapping:
                beam_current = ds["current"]
                beam_current_coords_to_apply = {
                    k: v for k, v in coords_to_apply.items() if k in beam_current.dims
                }
                beam_current = beam_current.assign_coords(beam_current_coords_to_apply)
                bc_units = beam_current.attrs.get("units", "mA")
                beam_current.attrs = {}
                da = da / beam_current
                units["spectrum"] = units["spectrum"] + f"/{bc_units}"
            else:
                analysis_warning(
                    "Could not determine the beam current to normalise data by",
                    "warning",
                    "Missing I0 data",
                )

        # Rename to peaks conventions using _manipulator_name_conventions and _analyser_name_conventions
        dim_names_to_update = {}
        names_to_check = cls._manipulator_name_conventions.copy()
        names_to_check.update(cls._analyser_name_conventions.copy())
        for peaks_name, i05_names in names_to_check.items():
            if isinstance(i05_names, str):
                if i05_names in dims:
                    dim_names_to_update[i05_names] = peaks_name
                    units[peaks_name] = units.pop(i05_names)
            elif isinstance(i05_names, list):
                for i05_name in i05_names:
                    if i05_name in dims:
                        dim_names_to_update[i05_name] = peaks_name
                        units[peaks_name] = units.pop(i05_name)
                        break
        da = da.rename(dim_names_to_update)

        # EDGE CASE - Handle problem with ana_polar data sometimes having non-monotonic points at end of range
        if "ana_polar" in da.dims:
            trim = 0
            while not (
                da.indexes["ana_polar"].is_monotonic_increasing
                or da.indexes["ana_polar"].is_monotonic_decreasing
            ):
                trim += 1
                da = da.isel(ana_polar=slice(0, -1))
            if trim:
                analysis_warning(
                    f"Non-monotonic angle values detected in ana_polar at end of travel. Trimmed {trim} data points.",
                    "warning",
                    "Removed non-monotonic data",
                )

        # Load array into memory if lazy loading not required
        if not (lazy or (lazy is None and da.size > opts.FileIO.lazy_size)):
            da = da.compute()

        # Add units where available
        da.name = "spectrum"
        da = da.pint.quantify(units)
        return da.squeeze()


class I05NanoFocussingMetadataModel(BaseMetadataModel):
    """Model to store metadata for Diamond Light Source Nano-ARPES focussing optics metadata."""

    OSAx: Optional[Union[str, Quantity]] = None
    OSAy: Optional[Union[str, Quantity]] = None
    OSAz: Optional[Union[str, Quantity]] = None
    ZPx: Optional[Union[str, Quantity]] = None
    ZPy: Optional[Union[str, Quantity]] = None
    ZPz: Optional[Union[str, Quantity]] = None


@register_loader
class I05NanoARPESLoader(I05ARPESLoader):
    _loc_name = "Diamond_I05_Nano-ARPES"
    _loc_description = "Nano-ARPES branch line of I05 beamline at Diamond Light Source"
    _loc_url = "https://www.diamond.ac.uk/Instruments/Structures-and-Surfaces/I05.html"
    _analyser_slit_angle = 0 * ureg.deg

    _manipulator_axes = ["polar", "tilt", "azi", "x1", "x2", "x3", "defocus"]
    _manipulator_name_conventions = {
        "polar": "smpolar",
        "tilt": None,
        "azi": "smazimuth",
        "x1": "smx",
        "x2": "smy",
        "x3": "smz",
        "defocus": "smdefocus",
    }
    _manipulator_sign_conventions = {
        # Nano-branch
        "azi": -1,
        "x1": -1,
        "x3": -1,
    }

    _analyser_sign_conventions = {
        "theta_par": -1,
        "deflector_parallel": -1,  # consistent with theta_par
    }

    _desired_dim_order = [
        "scan_no",
        "hv",
        "temperature_sample",
        "temperature_cryostat",
        "x3",
        "x2",
        "x1",
        "ana_polar",
        "polar",
        "tilt",
        "azi",
        "y_scale",
        "deflector_perp",
        "eV",
        "deflector_parallel",
        "theta_par",
    ]

    _analyser_name_conventions = {
        "deflector_perp": "ThetaY",
        "deflector_parallel": "ThetaX",
        "eV": ["energies", "kinetic_energy_center"],
        "theta_par": "angles",
        "hv": ["energy", "value"],
        "ana_polar": "analyser_polar_angle",
    }

    _hdf5_metadata_key_mappings = {
        "scan_command": "entry1/scan_command",
        "manipulator_polar": "entry1/instrument/manipulator/smpolar",
        "manipulator_tilt": "entry1/instrument/manipulator/smtilt",
        "manipulator_azi": "entry1/instrument/manipulator/smazimuth",
        "manipulator_x1": "entry1/instrument/manipulator/smx",
        "manipulator_x2": "entry1/instrument/manipulator/smy",
        "manipulator_x3": "entry1/instrument/manipulator/smz",
        "manipulator_defocus": "entry1/instrument/manipulator/smdefocus",
        "analyser_model": "FIXED_VALUE:Scienta DA30",
        "analyser_slit_width": [
            "entry1/instrument/analyser/entrance_slit_size",
            "entry1/instrument/analyser_total/entrance_slit_size",
        ],
        "analyser_slit_width_identifier": [
            "entry1/instrument/analyser/entrance_slit_setting",
            "entry1/instrument/analyser_total/entrance_slit_setting",
        ],
        "analyser_eV": [
            "entry1/instrument/analyser/energies",
            "entry1/instrument/analyser_total/energies",
            "entry1/instrument/analyser/kinetic_energy_center",
            "entry1/instrument/analyser_total/kinetic_energy_center",
        ],
        "analyser_step_size": [
            lambda f: (
                np.ptp(f["entry1/instrument/analyser/energies"])
                / (max(f["entry1/instrument/analyser/energies"].shape) - 1)
                if "entry1/instrument/analyser/energies" in f
                else None
            ),
            lambda f: (
                np.ptp(f["entry1/instrument/analyser_total/energies"])
                / (max(f["entry1/instrument/analyser_total/energies"].shape) - 1)
                if "entry1/instrument/analyser_total/energies" in f
                else None
            ),
            "entry1/instrument/analyser/kinetic_energy_step",
            "entry1/instrument/analyser_total/kinetic_energy_step",
        ],
        "analyser_PE": [
            "entry1/instrument/analyser/pass_energy",
            "entry1/instrument/analyser_total/pass_energy",
        ],
        "analyser_sweeps": [
            "entry1/instrument/analyser/number_of_iterations",
            "entry1/instrument/analyser/number_of_cycles",
            "entry1/instrument/analyser/number_of_frames",
            "entry1/instrument/analyser_total/number_of_iterations",
            "entry1/instrument/analyser_total/number_of_cycles",
            "entry1/instrument/analyser_total/number_of_frames",
        ],
        "analyser_dwell": [
            "entry1/instrument/analyser/time_for_frames",
            "entry1/instrument/analyser_total/time_for_frames",
        ],
        "analyser_lens_mode": [
            "entry1/instrument/analyser/lens_mode",
            lambda f: (
                f"FIXED_VALUE:{BaseHDF5DataLoader._extract_hdf5_value(f, 'entry1/instrument/analyser_total/lens_mode')}"
                f"_analyser_total"
                if "entry1/instrument/analyser_total/lens_mode" in f
                else None
            ),
        ],
        "analyser_acquisition_mode": [
            "entry1/instrument/analyser/acquisition_mode",
            "entry1/instrument/analyser_total/acquisition_mode",
        ],
        "analyser_eV_type": "FIXED_VALUE:kinetic",
        "analyser_deflector_parallel": None,
        "analyser_deflector_perp": None,
        "analyser_polar": [
            "entry1/instrument/analyser/analyser_polar_angle",
            "entry1/instrument/analyser_total/analyser_polar_angle",
        ],
        "temperature_sample": "entry1/sample/temperature",
        "temperature_cryostat": "entry1/sample/cryostat_temperature",
        "temperature_shield": "entry1/sample/shield_temperature",
        "temperature_setpoint": "entry1/sample/temperature_demand",
        "photon_hv": "entry1/instrument/monochromator/energy",
        "photon_polarisation": [
            "entry1/instrument/insertion_device/beam/final_polarisation_label",
            "entry1/instrument/insertion_device/beam/final_polarisation_label\n\t\t\t\t\t\t\t\t",
        ],
        "photon_exit_slit": "entry1/instrument/monochromator/exit_slit_size",
        "timestamp": "entry1/start_time",
        "focussing_OSAx": "entry1/instrument/order_sorting_aperture/osax",
        "focussing_OSAy": "entry1/instrument/order_sorting_aperture/osay",
        "focussing_OSAz": "entry1/instrument/order_sorting_aperture/osaz",
        "focussing_ZPx": "entry1/instrument/zone_plate/zpx",
        "focussing_ZPy": "entry1/instrument/zone_plate/zpy",
        "focussing_ZPz": "entry1/instrument/zone_plate/zpz",
    }

    _data_group_key_resolution_order = ["analyser", "analyser_total"]
    _core_data_key_resolution_order = ["analyser", "analyser_total"]

    _metadata_parsers = [
        "_parse_analyser_metadata",
        "_parse_manipulator_metadata",
        "_parse_photon_metadata",
        "_parse_temperature_metadata",
        "_parse_optics_metadata",
    ]  # List of metadata parsers to apply

    @classmethod
    def _load(cls, fpath, lazy, metadata, quiet, **kwargs):
        """Load the data from Diamond I05 Nano ARPES."""
        if fpath.split(".")[-1] == "zip":
            # Load the data using the SES loader for .zip files
            return cls.load(fpath, loc="SES", lazy=lazy, metadata=metadata)

        # Load the data
        data = super()._load(fpath, lazy, metadata, quiet, **kwargs)

        # Perform post processing on load if required
        if kwargs.pop("slant_correct", False):  # Perform the slant correction
            data = cls.slant_correct(data, kwargs.get("slant_factor", None))

        return data

    @classmethod
    def _parse_optics_metadata(cls, metadata_dict):
        """Parse the optics metadata from the metadata dictionary"""

        optics_metadata = I05NanoFocussingMetadataModel(
            OSAx=metadata_dict.get("focussing_OSAx"),
            OSAy=metadata_dict.get("focussing_OSAy"),
            OSAz=metadata_dict.get("focussing_OSAz"),
            ZPx=metadata_dict.get("focussing_ZPx"),
            ZPy=metadata_dict.get("focussing_ZPy"),
            ZPz=metadata_dict.get("focussing_ZPz"),
        )
        return {"_focussing": optics_metadata}, None

    @register_accessor(xr.DataArray)
    @staticmethod
    def slant_correct(data, slant_factor=None):
        """Function to remove a slant that was present in data obtained using the Scienta DA30 (9ES210) analyser at the nano
        branch of the I05 beamline at Diamond Light Source in 2021/22.

        Parameters
        ------------
        data : xarray.DataArray
            The data to be corrected.

        slant_factor : float, optional
            The slant factor correction (degrees/eV) to use. Defaults to 8/PE (where PE is the pass energy used).

        Returns
        ------------
        corrected_data : xarray.DataArray
            The corrected data.

        Examples
        ------------
        Example usage is as follows::

            from peaks import pks

            # Load the data, incorporating slant correction with a custom factor
            disp = pks.load('i05-1-13579.ibw', slant_correct=True, slant_factor=7.9)

            # Load data and apply slant correction afterwards
            disp = pks.load('i05-1-13579.ibw')
            disp_sc = disp.slant_correct()
        """

        # Set the slant factor to the default correction if it has not been supplied
        if slant_factor is None:
            slant_factor = 8 * ureg.deg / data.metadata.analyser.scan.PE.to("eV")
        elif not isinstance(slant_factor, pint.Quantity):
            slant_factor = slant_factor * ureg.deg / ureg.eV
        else:
            slant_factor = slant_factor.to("deg/eV")

        # Define the new angle mapping
        theta_par_values = data.theta_par.pint.to("deg") - (
            slant_factor.magnitude
            * (data.eV.pint.to("eV") - data.eV.pint.to("eV").median())
        )

        # Perform the interpolation onto the corrected grid
        corrected_data = (
            data.pint.dequantify()
            .interp({"theta_par": theta_par_values, "eV": data.eV})
            .pint.quantify()
        )

        # Update the analysis history
        corrected_data.history.add(
            "Slant correction for Diamond nano-ARPES data applied: {factor}".format(
                factor=slant_factor
            )
        )

        return corrected_data



"""Functions to load data from the Bloch beamline at MAX IV Laboratory."""

import pint_xarray

from peaks.core.fileIO.base_arpes_data_classes.base_ses_class import SESDataLoader
from peaks.core.fileIO.loc_registry import register_loader

ureg = pint_xarray.unit_registry


@register_loader
class BlochArpesLoader(SESDataLoader):
    _loc_name = "MAXIV_Bloch_A"
    _loc_description = "A branch (ARPES) of Bloch beamline at Max-IV"
    _loc_url = "https://www.maxiv.lu.se/beamlines-accelerators/beamlines/bloch/"
    _analyser_slit_angle = 0 * ureg("deg")

    _manipulator_name_conventions = {
        "polar": "P",
        "tilt": "T",
        "azi": "A",
        "x1": "X",
        "x2": "Y",
        "x3": "Z",
    }
    _manipulator_sign_conventions = {
        "polar": -1,
        "tilt": -1,
    }
    _analyser_sign_conventions = {
        "deflector_perp": -1,
    }
    _SES_metadata_units = {
        f"manipulator_{dim}": ("mm" if dim in ["x1", "x2", "x3"] else "deg")
        for dim in _manipulator_name_conventions.keys()
    }



import numpy as np
import xarray as xr

from peaks.core.fitting.fit import fit

QUICK_FIT_COMMON_DOC = """

Parameters
-----------
indepndent_var : str, optional
    The dimension corresponding to the indpendent variable. Must be specified if data has >1 dimension.
**kwargs : optional
    Additional keyword arguments to initialise parameter values.

Returns
---------
xarray.DataSet
    A DataSet containing the best-fit parameters, their uncertainties, and the :class:`lmfit.ModelResult` object.
"""


@xr.register_dataarray_accessor("quick_fit")
class QuickFit:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def _get_data_for_guess(self, model, independent_var, **kwargs):
        if independent_var is None and self._obj.ndim > 1:
            raise ValueError(
                f"Supplied data is {self._obj.ndim}-dimensional. Please specify the independent variable "
                f"with the argument `independent_var=_dim_`."
            )
        if self._obj.ndim > 1:
            dependent_vars = set(self._obj.dims) - set([independent_var])

            return self._obj.isel({dim: 0 for dim in dependent_vars}).squeeze().compute()
        else:
            return self._obj.compute()

    def _get_percentile_data(self, data, percentile=10, region="start"):
        """Get the data below or above a certain percentile of the dimension range

        Parameters
        -----------
        data : xarray.DataArray
            The data to extract the percentile from.
        percentile : int, optional
            The percentile to use for the cutoff. Defaults to 10.
        region : str, optional
            The region of the data to use for the fit. Options are 'start' or 'end'. Defaults to 'start'.

        Returns
        --------
        xarray.DataArray
            The data below or above the percentile cutoff.
        """

        dim = data.dims[0]
        cutoff = np.percentile(data[dim].data, percentile)
        if region == "start":
            return data.sel({dim: slice(None, cutoff)})
        elif region == "end":
            return data.sel({dim: slice(cutoff, None)})
        else:
            raise ValueError("Region must be 'start' or 'end'")

    def linear(self, independent_var=None, **kwargs):
        """Quick fit to a linear model"""
        from .models import LinearModel

        data_for_guess = self._get_data_for_guess(
            LinearModel(), independent_var, **kwargs
        )
        params = LinearModel().guess(data_for_guess, **kwargs)
        return fit(self._obj, LinearModel(), params, independent_var)

    linear.__doc__ = linear.__doc__ + QUICK_FIT_COMMON_DOC

    def poly(self, degree=3, independent_var=None, **kwargs):
        """Quick fit to a polynomial model

        Parameters
        -----------
        independent_var : str, optional
            The dimension corresponding to the independent variable. Must be specified if data has >1 dimension.
        degree : int, optional
            The degree of the polynomial model. Defaults to 3.
        **kwargs : optional
            Additional keyword arguments to initialise parameter values.

        Returns
        ---------
        xarray.DataSet
            A DataSet containing the best-fit parameters, their uncertainties, and the :class:`lmfit.ModelResult` object.
        """
        from .models import PolynomialModel

        data_for_guess = self._get_data_for_guess(
            PolynomialModel(degree=degree), independent_var
        )
        params = PolynomialModel(degree=degree).guess(data_for_guess, **kwargs)
        return fit(self._obj, PolynomialModel(degree=degree), params, independent_var)

    def _peak_model(self, peak, independent_var, **kwargs):
        """Quick fit to a Gaussian model"""
        from .models import LinearModel

        if peak == "gaussian":
            from .models import GaussianModel

            peak_model = GaussianModel()
        elif peak == "lorentzian":
            from .models import LorentzianModel

            peak_model = LorentzianModel()

        model = peak_model + LinearModel()
        data_for_guess = self._get_data_for_guess(model, independent_var, **kwargs)
        params = model.make_params(**kwargs)
        params.update(peak_model.guess(data_for_guess, **kwargs))
        params.update(
            LinearModel().guess(self._get_percentile_data(data_for_guess), **kwargs)
        )
        return fit(self._obj, model, params, independent_var)

    def gaussian(self, independent_var=None, **kwargs):
        """Quick fit to a Gaussian model"""
        return self._peak_model("gaussian", independent_var, **kwargs)

    gaussian.__doc__ = gaussian.__doc__ + QUICK_FIT_COMMON_DOC

    def lorentzian(self, independent_var=None, **kwargs):
        """Quick fit to a Lorentzian model"""
        return self._peak_model("lorentzian", independent_var, **kwargs)

    lorentzian.__doc__ = lorentzian.__doc__ + QUICK_FIT_COMMON_DOC



"""Functions used to extract selections of data."""

import copy

import numpy as np
import pint_xarray  # noqa: F401
import xarray as xr
from matplotlib.path import Path

from peaks.core.utils.interpolation import (
    _fast_bilinear_interpolate,
    _fast_bilinear_interpolate_rectilinear,
    _is_linearly_spaced,
)
from peaks.core.utils.misc import dequantify_quantify_wrapper

ureg = pint_xarray.unit_registry


def drop_nan_borders(data):
    """
    Trims the edges of an :class:`xarray.DataArray` or :class:`xarray.Dataset` that are all NaN values.

    Parameters
    ----------
    data : xarray.DataArray or xarray.Dataset
        The data to trim.

    Returns
    --------
    trimmed_data : xarray.DataArray or xarray.Dataset
        Trimmed :class:`xarray` object with NaN edges removed.

    Examples
    --------
    Example usage is as follows::

            import peaks as pks

            # Load a data file
            data = load("data.nxs")

            # Trim all zero edges from the data
            trimmed_data = data.drop_nan_borders()
    """

    # Iterate over each dimension to trim NaN edges
    for dim in data.dims:
        # Create a boolean mask where data is not NaN along the dimension
        not_nan = ~np.isnan(data).any(dim=[d for d in data.dims if d != dim])

        # Find the first and last index where the data is not all NaN
        non_nan_indices = np.where(not_nan)[0]
        if non_nan_indices.size > 0:
            start = non_nan_indices[0]
            end = non_nan_indices[-1] + 1
            data = data.isel({dim: slice(start, end)})

    return data


def _drop_nan_borders_2D(data):
    """Function to drop rows and columns containing only NaNs from a 2D array."""
    # Drop rows and columns containing only NaNs
    # Depreciate in future in favour of drop_nan_borders
    rows_to_keep = ~np.all(np.isnan(data), axis=1)
    cols_to_keep = ~np.all(np.isnan(data), axis=0)
    return data[rows_to_keep, :][:, cols_to_keep]


def drop_zero_borders(data):
    """
    Trims the edges of an :class:`xarray.DataArray` or :class:`xarray.Dataset` that are all zero values.

    Parameters
    ----------
    data : xarray.DataArray or xarray.Dataset
        The data to trim.

    Returns
    --------
    trimmed_data : xarray.DataArray or xarray.Dataset
        Trimmed :class:`xarray` object with NaN edges removed.

    Examples
    --------
    Example usage is as follows::

            import peaks as pks

            # Load a data file
            data = load("data.nxs")

            # Trim all zero edges from the data
            trimmed_data = data.drop_zero_borders()
    """

    # Iterate over each dimension to trim zero edges
    for dim in data.dims:
        # Create a boolean mask where data is not zero along the dimension
        not_zero = (data != 0).any(dim=[d for d in data.dims if d != dim])

        # Find the first and last index where the data is not all zero
        non_zero_indices = np.where(not_zero)[0]
        if non_zero_indices.size > 0:
            start = non_zero_indices[0]
            end = non_zero_indices[-1] + 1
            data = data.isel({dim: slice(start, end)})

    return data


@dequantify_quantify_wrapper
def DC(data, coord="eV", val=0, dval=0, ana_hist=True):
    """General function to extract DCs from data along any coordinate.

    Parameters
    ------------
    data : xarray.DataArray
        The data to extract a DC from.

    coord : str, optional
        Coordinate to extract DC at. Defaults to eV.

    val : float, list, numpy.ndarray, tuple, optional
        DC value(s) to select. If tuple, must be in the format (start, end, step). Defaults to 0.

    dval : float, optional
        Integration range (represents the total range, i.e. integrates over +/- dval/2). Defaults to 0.

    ana_hist : bool, optional
        Defines whether the function appends information to the analysis history metadata. Defaults to True.

    Returns
    ------------
    dc : xarray.DataArray
        Extracted DC(s).

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        # Extract an EDC at theta_par = 3.5 +/- 0.25
        DC1 = disp.DC(coord='theta_par', val=3.5, dval=0.5)

        # Extract an MDC at eV = -0.5 +/- 0.1
        DC2 = disp.DC('eV', -0.5, 0.2)

        # Extract MDCs at eV = -0.5 +/- 0.1 and -0.4 +/- 0.1
        DC3 = disp.DC('eV', [-0.5, -0.4], 0.2)

        # Extract MDCs between eV = -0.2 and 0.1 in steps of 0.05 with +/- 0.01 integrations
        DC4 = disp.DC('eV', (-0.2, 0.1, 0.05), 0.02)

    """

    # If val is a 3 element tuple of the format (start, end, step), make val an numpy.ndarray of the relevant values
    if isinstance(val, tuple):
        if len(val) == 3:
            delta = 0.000000001  # small value to add to end so that values include end number (if appropriate)
            val = np.arange(val[0], val[1] + delta, val[2])
        else:
            raise Exception("Tuple argument must be in the format (start, end, step).")

    # Ensure val is of type list
    if isinstance(val, np.ndarray):
        val = list(val)
    elif not isinstance(val, list):
        val = [val]

    # Convert window to pixels
    num_pixels = int((dval / abs(data[coord].data[1] - data[coord].data[0])) + 1)

    # Extract single pixel DC(s)
    dc = data.sel({coord: val}, method="nearest")

    # Apply extra binning if required
    if num_pixels > 1:
        for i in dc[coord]:
            dc.loc[{coord: i}] = data.sel(
                {coord: slice(i - dval / 2, i + dval / 2)}
            ).mean(coord, keep_attrs=True)

    # If returning a single DC, want to remove the non-varying coordinate as a dimension
    try:
        dc = dc.squeeze(coord)
    except ValueError:
        pass

    # Update the analysis history if ana_hist is True (will be False when DC is called from e.g. EDC, MDC, FS)
    if ana_hist:
        hist = "DC(s) extracted, integration window: " + str(dval)
        dc = dc.history.assign(hist)

    return dc


def MDC(data, E=0, dE=0):
    """Extract MDCs (i.e. slices at constant energy) from data. Broadcasts to higher dimensions as necessary.

    Parameters
    ------------
    data : xarray.DataArray
        The dispersion to extract an MDC from.

    E : float, list, numpy.ndarray, tuple, optional
        Energy (or energies) of MDC(s) to extract. If tuple, must be in the format (start, end, step). Defaults to 0.

    dE : float, optional
        Integration range (represents the total range, i.e. integrates over +/- dE/2). Defaults to 0.

    Returns
    ------------
    mdc : xarray.DataArray
        Extracted MDC(s).

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # Load data
        disp = pks.load('disp.ibw')
        FS1 = pks.load('FS.ibw').k_convert()  # Convert to k-space and BE

        # Extract single MDCs from the dispersion
        MDC1 = disp.MDC(E=-1.2, dE=0.01)  # at eV = -1.2 +/- 0.005
        MDC2 = disp.MDC(55.6, 0.06)  # at eV = 55.6 +/- 0.03
        MDC3 = disp.MDC()  # a single non-integrated MDC at eV value closest to 0

        # Extract multiple MDCs from the dispersion
        MDC4 = disp.MDC([55.6, 55.7], 0.06)  # at eV = 55.6 & 55.7, +/- 0.03
        MDC5 = disp.MDC((-0.2, 0.1, 0.05), 0.02)  # eV = -0.2 and 0.1 in steps of 0.05 with +/- 0.01 integrations

        # Extract a Fermi surface, data already in binding energy
        FS_map = FS1.MDC()  # No integration required - defaults are good!
        FS_map2 = FS1.MDC(dE=0.02)  # Integration over +/- 0.01 eV
    """

    # Call function to extract relevant MDC from dispersion
    mdc = data.DC(coord="eV", val=E, dval=dE, ana_hist=False)

    # Update the analysis history
    hist = "MDC(s) extracted, integration window: " + str(dE)
    mdc = mdc.history.assign(hist)

    return mdc


def EDC(data, k=0, dk=0):
    """Extract EDCs from data.

    Parameters
    ------------
    data : xarray.DataArray
        The dispersion to extract an EDC from.

    k : float, list, numpy.ndarray, tuple, optional
        k or theta_par value(s) of EDC(s) to extract. If tuple, must be in the format (start, end, step). Defaults to 0.

    dk : float, optional
        Integration range (represents the total range, i.e. integrates over +/- dk/2). Defaults to 0.

    Returns
    ------------
    edc : xarray.DataArray
        Extracted EDC(s).

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        # Extract an EDC at k (or theta_par) = 0.5 +/- 0.005
        EDC1 = disp.EDC(k=0.5, dk=0.01)

        # Extract an EDC at k (or theta_par) = -0.2 +/- 0.03
        EDC2 = disp.EDC(-0.2, 0.06)

        # Extract a single non-integrated EDC at k (or theta_par) value closest to 0
        EDC3 = disp.EDC()

        # Extract EDCs at k (or theta_par) = -0.2 +/- 0.03 and -0.1 +/- 0.03
        EDC4 = disp.EDC([-0.2, -0.1], 0.06)

        # Extract EDCs between k (or theta_par) = 0.7 and 1.2 in steps of 0.1 with +/- 0.01 integrations
        EDC5 = disp.EDC((0.7, 1.2, 0.1), 0.02)

    """

    # Work out correct variable for dispersive direction (i.e. is data in angle or k-space)
    coords = list(data.dims)
    coords.remove("eV")
    coord = coords[-1]  # Should always be the last one if data loading is consistent

    # Call function to extract relevant EDC from dispersion
    edc = data.DC(coord=coord, val=k, dval=dk, ana_hist=False)

    # Update the analysis history
    hist = "EDC(s) extracted, integration window: " + str(dk)
    edc = edc.history.assign(hist)

    return edc


def DOS(data):
    """Integrate over all but the energy axis to return the best approximation to the DOS possible from the data.

    Parameters
    ------------
    data : xarray.DataArray
        Data to extract DOS from.

    Returns
    ------------
    dos : xarray.DataArray
        Extracted DOS.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        FM = load('FM.zip')

        # Extract DOS of a dispersion
        disp_DOS = disp.DOS()

        # Extract DOS of a Fermi map
        FM_DOS = FM.DOS()

    """

    # Get relevant dimensions to integrate over
    int_dim = list(filter(lambda i: i != "eV", data.dims))

    # Calculate the DOS
    dos = data.mean(int_dim, keep_attrs=True)

    # Update the analysis history
    hist = "Integrated along axes: " + str(int_dim)
    dos = dos.history.assign(hist)

    return dos


def tot(data, spatial_int=False):
    """Integrate spatial map data over all non-spatial (energy and angle/k) or all spatial dimensions.

    Parameters
    ------------
    data : xarray.DataArray
        Spatial map data.

    spatial_int : bool, optional
        Determines whether integration is performed over spatial or non-spatial dimensions. Defaults to False.

    Returns
    ------------
    data_tot : xarray.DataArray
        The integrated data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM.ibw')

        # Extract energy and angle integrated spatial map
        SM_int = SM.tot()

        # Extract spatially integrated dispersion
        SM_int_spatial = SM.tot(spatial_int=True)

    """

    # Integrate over spatial dimensions
    if spatial_int:
        data_tot = data.mean(["x1", "x2"], keep_attrs=True)
        hist = "Integrated along axes: " + str(["x1", "x2"])

    # Integrate over non-spatial dimensions
    else:
        # Get relevant dimensions to integrate over
        int_dim = list(filter(lambda n: n != "x1" and n != "x2", data.dims))
        hist = "Integrated along axes: " + str(int_dim)
        data_tot = data.mean(int_dim, keep_attrs=True)

    # Update the analysis history
    data_tot = data_tot.history.assign(hist)

    return data_tot


@dequantify_quantify_wrapper
def radial_cuts(data, num_azi=361, num_points=200, radius=2, **centre_kwargs):
    """Extract radial cuts of a Fermi surface slice or cube as a function of azimuthal angle, about some central point.

    Parameters
    ------------
    data : xarray.DataArray
        Data to extract radial cuts from.

    num_azi : float, optional
        Number of evenly spaced azi values between 0 and 360 degrees to take radial cuts. Defaults to 361.

    num_points : float, optional
        Number of evenly spaced points to sample along a cut. Defaults to 200.

    radius : float, optional
        Maximum radius to take cuts up to. Defaults to 2.

    **centre_kwargs : float, optional
        Used to define centre of rotations in the format dim = coord, e.g. k_par = 1.2 sets the k_par centre as 1.2.
        Default centre of rotation is (0, 0).

    Returns
    ------------
    data_to_return : xarray.DataArray
        Radial cuts against azi.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        FM = pks.load('FM.zip')

        FS1 = FM.MDC(E=-0, dE=0.01)

        # Extract radial cuts (radius = 15) at azi values in 2 degree increments, using a centre of
        # rotation (theta_par, ana_polar) = (-2, -5)
        FS1_radial_cuts_1 = FS1.radial_cuts(num_azi=181, num_points=300, radius=15, theta_par=-2, ana_polar=-5)

        # Extract radial cuts (radius = 2) at azi values in 1 degree increments, using a centre of
        # rotation (coord_1, coord_2) = (0, 0)
        FS1_radial_cuts_2 = FS1.radial_cuts()

    """

    angle_dims = list(set(data.dims) - {"eV"})

    # Check remaining data is 2D
    if len(angle_dims) != 2:
        raise Exception(
            "Radial cuts can only be taken on data with two angle/k-space dimensions,"
            "with optionally an additional energy dimension."
        )

    # Define the coordinate system
    ang0_coord = angle_dims[0]
    ang1_coord = angle_dims[1]

    # Check for user-defined centre of rotations
    ang0_centre = centre_kwargs.get(ang0_coord)
    if not ang0_centre:
        ang0_centre = 0
    ang1_centre = centre_kwargs.get(ang1_coord)
    if not ang1_centre:
        ang1_centre = 0

    # Define coordinates to be sampled
    azi_angles = np.linspace(0, 360, num_azi)
    k_values = np.linspace(0, radius, num_points)

    # Calculate the values for interpolation
    ang0_values = np.linspace(
        0 + ang0_centre,
        (np.cos(np.radians(azi_angles)) * radius) + ang0_centre,
        num_points,
    )
    ang1_values = np.linspace(
        0 + ang1_centre,
        (np.sin(np.radians(azi_angles)) * radius) + ang1_centre,
        num_points,
    )

    # Check if we have a rectilinear grid to determine the interpolation function
    if _is_linearly_spaced(data[ang0_coord].data) and _is_linearly_spaced(
        data[ang1_coord].data
    ):
        interpolation_fn = _fast_bilinear_interpolate_rectilinear
    else:
        interpolation_fn = _fast_bilinear_interpolate

    # Do the interpolation, broadcasting over energy dimension if required
    interpolated_data = xr.apply_ufunc(
        interpolation_fn,
        ang0_values,
        ang1_values,
        data[ang0_coord].data,
        data[ang1_coord].data,
        data,
        dask="parallelized",
        input_core_dims=[
            ["k", "azi"],
            ["k", "azi"],
            [ang0_coord],
            [ang1_coord],
            [ang0_coord, ang1_coord],
        ],
        output_core_dims=[["k", "azi"]],
        output_dtypes=[data.dtype],
        vectorize=True,
        dask_gufunc_kwargs={"allow_rechunk": True},
    )

    # Update co-ordinates
    interpolated_data["k"] = k_values
    interpolated_data["azi"] = azi_angles

    # Update attributes and analysis history
    interpolated_data.attrs = copy.deepcopy(data.attrs)
    interpolated_data.history.add("Radial cuts taken as a function of azi")

    return interpolated_data


@dequantify_quantify_wrapper
def extract_cut(data, start_point, end_point, num_points=None):
    """Extract cut between two end-points, e.g. dispersion or MDC from a Fermi map data cube.

    Parameters
    ------------
    data : xarray.DataArray
        Data to extract radial cuts from.

    start_point : dict
        Dictionary containing the coordinates of the start of the desired cut to extract.

    end_point : dict
        Dictionary containing the coordinates of the end of the desired cut to extract.

    num_points : int, optional
        Number of points to sample along the cut. Defaults to None, in which case spacing is determined
        based on the original data.

    Returns
    ------------
    data_to_return : xarray.DataArray
        Extracted cut

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        FM = pks.load('FM.zip')

        # Extract cut from (0, 0) to (15, 12) with 100 points
        cut = FM.extract_dispersion({'theta_par': 0, 'polar': 0}, {'theta_par': 15, 'polar': 12}, num_points=100)

    """

    # Check the start and end points specification
    if set(start_point.keys()) != set(end_point.keys()):
        raise ValueError(
            "Start and end points must be specified as dictionaries `start_point={dim0: start_point, dim1: start_point}`"
            " and `end_point={dim0: end, dim1: end_point}`, with the same dimensions specified in both."
        )
    if not all([key in data.dims for key in start_point.keys()]):
        raise ValueError(
            "Ensure the dimensions specified in the start and end points are present in the data."
        )

    # Define the coordinate system
    dim0, dim1 = start_point.keys()

    # Calculate projection vector along the cut from the start point
    start_coord0, start_coord1 = start_point[dim0], start_point[dim1]
    end_coord0, end_coord1 = end_point[dim0], end_point[dim1]
    distance = np.sqrt(
        (end_coord0 - start_coord0) ** 2 + (end_coord1 - start_coord1) ** 2
    )
    # If num_points is not specified, calculate the number of points based on the step size in the original data
    if num_points is None:
        step_size = np.sqrt(
            (data[dim0].data[1] - data[dim0].data[0]) ** 2
            + (data[dim1].data[1] - data[dim1].data[0]) ** 2
        )
        num_points = int(np.ceil(distance / step_size))
    projection = np.linspace(0, distance, num_points)

    # Calculate the values for interpolation
    ang0_values = np.linspace(start_point[dim0], end_point[dim0], num_points)
    ang1_values = np.linspace(start_point[dim1], end_point[dim1], num_points)

    # Check if we have a rectilinear grid to determine the interpolation function
    if _is_linearly_spaced(data[dim0].data) and _is_linearly_spaced(data[dim1].data):
        interpolation_fn = _fast_bilinear_interpolate_rectilinear
    else:
        interpolation_fn = _fast_bilinear_interpolate

    # Do the interpolation, broadcasting over remaining dimensions if required
    interpolated_data = xr.apply_ufunc(
        interpolation_fn,
        ang0_values,
        ang1_values,
        data[dim0].data,
        data[dim1].data,
        data,
        input_core_dims=[
            ["proj"],
            ["proj"],
            [dim0],
            [dim1],
            [dim0, dim1],
        ],
        output_core_dims=[["proj"]],
        output_dtypes=[data.dtype],
        vectorize=True,
        dask="parallelized",
        dask_gufunc_kwargs={"allow_rechunk": True},
    )

    # Update co-ordinates
    interpolated_data["proj"] = projection

    # Update attributes and analysis history
    interpolated_data.attrs = copy.deepcopy(data.attrs)
    interpolated_data.history.add(
        f"Slice extracted from data from {start_point} to {end_point}. "
        f"Data returned vs. the projected distance."
    )

    interpolated_data = interpolated_data.pint.quantify()

    return interpolated_data


def mask_data(data, ROI, return_integrated=True):
    """This function applies a polygon region of interest (ROI) as a mask to multidimensional data. By default, the
    function will then extract the mean over the two dimensions defined by the ROI. For a rectangular ROI, this is
    equivalent to a simple .sel over those dimensions followed by a mean, but an arbitrary polygon can be used to define
    the ROI.

    Parameters
    ------------
    data : xarray.DataArray
        The multidimensional data to apply the ROI selection to.

    ROI : dict
        A dictionary of two lists which contains the vertices of the polygon for the ROI definition, in the form
        {'dim1': [pt1, pt2, pt3, ...], 'dim2'=[pt1', pt2', pt3', ...]}. As many points can be specified as required,
        but this should be given with the same number of points for each dimension.

    return_integrated : bool, optional
        Whether to mean the data confined within ROI region over the ROI dimensions, or instead return the masked data.
        Defaults to True.

    Returns
    ------------
    ROI_selected_data : xarray.DataArray
        The input data with the ROI applied as a mask, and (if return_integrated=True) the mean taken over those
        remaining dimensions.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM.ibw')

        # Define ROI used to mask data
        ROI = {'theta_par': [-8, -5.5, -3.1, -5.6], 'eV': [95.45, 95.45, 95.77, 95.77]}

        # Extract SM consisting of the integrated spectral weight confined within the ROI
        ROI_SM = SM.mask_data(ROI)

        # Extract SM consisting of the input data with the ROI applied as a mask
        ROI_SM = SM.mask_data(ROI, return_integrated=False)

    """

    # Check function has been fed with suitable dictionary for ROI generation
    err_str = (
        "ROI must be a dictionary containing two entries for the relevant axes. Each of these entries should "
        "be a list of the vertices of the polygon for the labelled axis of the ROI. These must be of equal "
        "length for the two axes."
    )
    if not isinstance(ROI, dict) or len(ROI) != 2:
        raise Exception(err_str)

    else:  # Seems correct format
        dims = list(ROI)  # Determine relevant dimensions for ROI

        # Check lengths match
        if len(ROI[dims[0]]) != len(ROI[dims[1]]):
            raise Exception(err_str)

    # Define ROI_path to make a polygon path defining the ROI
    ROI_path = []
    for i in range(len(ROI[dims[0]])):
        ROI_path.append((ROI[dims[0]][i], ROI[dims[1]][i]))
    p = Path(ROI_path)  # Make a polygon defining the ROI

    # Restrict the data cube down to the minimum possible size (making this a copy to avoid overwriting problems)
    data_bounded = data.sel(
        {
            dims[0]: slice(min(ROI[dims[0]]), max(ROI[dims[0]])),
            dims[1]: slice(min(ROI[dims[1]]), max(ROI[dims[1]])),
        }
    ).copy(deep=True)

    # Broadcast coordinate data
    b, dim0 = xr.broadcast(data_bounded, data_bounded[dims[0]])
    b, dim1 = xr.broadcast(data_bounded, data_bounded[dims[1]])

    # Convert coordinate data into a format for passing to matplotlib path function for identifying which points are
    # in the relevant ROI
    points = np.vstack((dim0.data.flatten(), dim1.data.flatten())).T

    # Check which of these points fall within our ROI
    grid = p.contains_points(points, radius=0.01)

    # Reshape to make a data mask
    mask = grid.reshape(data_bounded.shape)

    if return_integrated:  # Data should be averaged over the ROI dimensions
        ROI_selected_data = data_bounded.where(mask).mean(dims, keep_attrs=True)
        hist = (
            "Data averaged over region of interest defined by polygon with vertices: "
            + str(ROI)
        )
    else:  # Masked data to be returned
        ROI_selected_data = data_bounded.where(mask)

        # Trim any rows or columns of only NaNs
        try:  # The top method works for 2D, but seems to fail for higher dimensions
            ROI_selected_data = _drop_nan_borders_2D(ROI_selected_data)
        except Exception:
            ROI_selected_data = drop_nan_borders(ROI_selected_data)

        hist = (
            "Data masked by region of interest defined by polygon with vertices: "
            + str(ROI)
        )

    # Update analysis history
    ROI_selected_data = ROI_selected_data.history.assign(hist)

    return ROI_selected_data


def disp_from_hv(da, hv):
    """Function to extract a dispersion at a given hv from an hv scan, correcting for the kinetic energy offsets
    (KE_delta) that arise from using the hv scan loading method.

    Parameters
    ------------
    da : xarray.DataArray
        The hv scan extract a single dispersion from.

    hv : float
         The photon energy at which to extract a dispersion at.

    Returns
    ------------
    hv_disp : xarray.DataArray
        The extracted dispersion.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        hv_scan = load('i05-1-54321.nxs')

        # Extract a dispersion from the hv_scan at 79 eV
        disp_79eV = hv_scan.disp_from_hv(79)

    """

    # Ensure the inputted data is an hv scan
    if "hv" not in da.dims:
        raise Exception(
            "The scan type of the inputted data is incompatible with disp_from_hv, which only extracts a "
            "single dispersion from an hv scan."
        )

    # Extract the relevant hv slice
    hv_scan = da.sel(hv=hv, method="nearest")

    # If the inputted data is in binding energy, we are done
    if "binding" in da.metadata.analyser.scan.eV_type.lower():
        return hv_scan

    # Rescale eV axis to get the correct kinetic energy
    orig_units = hv_scan.eV.units
    if isinstance(hv_scan.KE_delta.data, np.ndarray):
        hv_scan["eV"] = hv_scan.eV.data + hv_scan.KE_delta.data
    else:
        hv_scan["eV"] = (
            hv_scan.eV.data + hv_scan.KE_delta.pint.to(orig_units).pint.dequantify().data
        )  # Handle unit conversion
        hv_scan = hv_scan.pint.quantify(eV=orig_units)  # Add the units back

    # Delete the now redundant hv and KE_delta coordinates
    del hv_scan["hv"]
    del hv_scan["KE_delta"]

    # Update the hv attribute
    hv_scan.metadata.photon.set("hv", float(hv), add_history=False)

    # Update analysis history
    hv_scan = hv_scan.history.assign(
        "Dispersion extracted from hv scan at hv={hv} eV".format(hv=hv)
    )

    return hv_scan
